import re
from typing import cast

from bs4 import BeautifulSoup, Tag

from labels.model.file import LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.dotnet.package_builder import new_dotnet_package
from labels.parsers.cataloger.utils import get_enriched_location

PACKAGE = re.compile(r".+\\packages\\(?P<package_info>[^\s\\]*)\\.+")
DEP_INFO = re.compile(r"(?P<package_name>.*?)\.(?P<version>\d+[^\s]*)$")


PACKAGE_REFERENCE_ELEMENT = "packagereference"
PRIVATE_ASSETS_ELEMENT = "privateassets"
REFERENCE_ELEMENT = "reference"
HINT_PATH_ELEMENT = "hintpath"


def parse_csproj(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = BeautifulSoup(reader.read_closer.read(), features="html.parser")

    packages = _collect_packages(file_content, reader)

    return packages, []


def _collect_packages(root: BeautifulSoup, reader: LocationReadCloser) -> list[Package]:
    packages: list[Package] = []

    package_references: list[Tag] = root.find_all(PACKAGE_REFERENCE_ELEMENT, recursive=True)
    for raw_package in package_references:
        package_name = cast("str | None", raw_package.get("include"))
        version = cast("str | None", raw_package.get("version"))

        is_dev = _is_dev_dependency(raw_package)
        new_location = get_enriched_location(
            reader.location,
            line=raw_package.sourceline,
            is_transitive=False,
            is_dev=is_dev,
            package_manager=PackageManager.NUGET,
        )

        package = new_dotnet_package(package_name, version, new_location)
        if package:
            packages.append(package)

    packages.extend(_collect_package_from_reference_dependencies(root, reader))

    return packages


def _is_dev_dependency(pkg: Tag) -> bool:
    checking_attr = "all"

    element = pkg.find(PRIVATE_ASSETS_ELEMENT)
    if isinstance(element, Tag):
        return element.get_text(strip=True).lower() == checking_attr

    attr = pkg.get(PRIVATE_ASSETS_ELEMENT)
    if isinstance(attr, str):
        return attr.strip().lower() == checking_attr

    return False


def _collect_package_from_reference_dependencies(
    raw_package: Tag, reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []
    references: list[Tag] = raw_package.find_all(REFERENCE_ELEMENT, recursive=True)

    for reference in references:
        package = _extract_package_from_reference(reference, reader)
        if package:
            packages.append(package)

    return packages


def _extract_package_from_reference(reference: Tag, reader: LocationReadCloser) -> Package | None:
    package = _extract_from_hint_path(reference, reader)
    if package:
        return package

    return _extract_from_include_attribute(reference, reader)


def _extract_from_hint_path(reference: Tag, reader: LocationReadCloser) -> Package | None:
    dll_path = reference.find(HINT_PATH_ELEMENT)
    if not isinstance(dll_path, Tag):
        return None

    dll_package = PACKAGE.match(dll_path.text)
    if not dll_package:
        return None

    package_info = DEP_INFO.match(dll_package.group("package_info"))
    if not package_info:
        return None

    new_location = get_enriched_location(
        reader.location,
        line=dll_path.sourceline,
        is_transitive=False,
        package_manager=PackageManager.NUGET,
    )

    name: str = package_info.group("package_name").lower()
    version: str = package_info.group("version")

    return new_dotnet_package(name, version, new_location)


def _extract_from_include_attribute(raw_package: Tag, reader: LocationReadCloser) -> Package | None:
    include = raw_package.get("include")
    if not isinstance(include, str):
        return None

    include_info = include.replace(" ", "").split(",")

    package_name = include_info[0].strip()
    version = _get_version(include_info)

    if not package_name or not version:
        return None

    new_location = get_enriched_location(
        reader.location,
        line=raw_package.sourceline,
        is_transitive=False,
        package_manager=PackageManager.NUGET,
    )

    return new_dotnet_package(package_name, version, new_location)


def _get_version(include_info: list[str]) -> str | None:
    return next(
        (
            package_info.lstrip("Version=")
            for package_info in include_info
            if package_info.startswith("Version=")
        ),
        None,
    )



import re
from pathlib import Path
from typing import TypeAlias

from fluidattacks_core.semver.match_versions import match_version_ranges

from labels.model.file import DependencyType, Location
from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import (
    LOGGER,
    get_enriched_location,
)
from labels.parsers.collection.json import parse_json_with_tree_sitter

DependencyRequirements: TypeAlias = dict[str, str]

DependencyMap: TypeAlias = dict[str, DependencyRequirements]


X_RANGE_PATTERN = re.compile(r"^\d+(?:\.\d+)?\.x\Z")
BARE_NUMBER_PATTERN = re.compile(r"^\d+(?:\.\d+)?\Z")
HYPHEN_RANGE_PATTERN = re.compile(r"^(\S+)\s+-\s+(\S+)\Z")
OPERATOR_PATTERN = re.compile(r"^([><=!]+)(.+)\Z")


def _bare_number_to_bounds(token: str) -> tuple[str, str]:
    parts = token.split(".")
    upper_parts = list(parts)
    upper_parts[-1] = str(int(upper_parts[-1]) + 1)
    return f"{token}.0", ".".join(upper_parts) + ".0"


def _x_range_to_bounds(x_token: str) -> tuple[str, str]:
    return _bare_number_to_bounds(x_token.removesuffix(".x"))


def _range_token_to_bounds(token: str) -> tuple[str, str]:
    if X_RANGE_PATTERN.match(token):
        return _x_range_to_bounds(token)
    return _bare_number_to_bounds(token)


def _is_range_token(token: str) -> bool:
    return bool(X_RANGE_PATTERN.match(token) or BARE_NUMBER_PATTERN.match(token))


def _convert_hyphen_range(lower: str, upper: str) -> str:
    upper_parts = upper.split(".")
    if len(upper_parts) < 3 and upper_parts[-1].isdigit():
        upper_parts[-1] = str(int(upper_parts[-1]) + 1)
        return f">={lower} <{'.'.join(upper_parts)}.0"
    return f">={lower} <={upper}"


def _version_tuple(version: str) -> tuple[int, ...]:
    return tuple(int(s) for s in version.split(".") if s.isdigit())


def _bump_leftmost_nonzero(int_parts: list[int]) -> list[int]:
    for i, val in enumerate(int_parts):
        if val != 0:
            int_parts[i] = val + 1
            int_parts[i + 1 :] = [0] * (len(int_parts) - i - 1)
            return int_parts
    int_parts[-1] = 1
    return int_parts


def _implicit_upper_from_caret_tilde(token: str) -> str | None:
    if len(token) < 2 or token[0] not in ("^", "~"):
        return None

    prefix, version_str = token[0], token[1:]
    try:
        parts = [int(p) for p in version_str.split(".")]
    except ValueError:
        return None

    if prefix == "^":
        parts = _bump_leftmost_nonzero(parts)

    if prefix == "~":
        if len(parts) >= 2:
            parts[1] += 1
            parts[2:] = [0] * (len(parts) - 2)
        else:
            parts[0] += 1

    parts += [0] * max(0, 3 - len(parts))

    return ".".join(str(p) for p in parts)


def _is_tighter_upper(new_op: str, new_ver: str, cur_op: str, cur_ver: str) -> bool:
    new_key = (_version_tuple(new_ver), new_op.endswith("="))
    cur_key = (_version_tuple(cur_ver), cur_op.endswith("="))
    return new_key < cur_key


def _resolve_caret_tilde_range(caret_tilde_token: str, operators: list[str]) -> str:
    implicit_lower = caret_tilde_token[1:]
    implicit_upper = _implicit_upper_from_caret_tilde(caret_tilde_token)
    if implicit_upper is None:
        return " ".join([implicit_lower, *operators])

    lower_op = ">="
    lower = implicit_lower
    upper_op = "<"
    upper = implicit_upper

    for token in operators:
        if match := OPERATOR_PATTERN.match(token):
            op: str = match.group(1)
            ver: str = match.group(2)

            if op.startswith(">") and _version_tuple(ver) > _version_tuple(lower):
                lower_op = op
                lower = ver
            elif op.startswith("<") and _is_tighter_upper(op, ver, upper_op, upper):
                upper_op = op
                upper = ver

    return f"{lower_op}{lower} {upper_op}{upper}"


def _normalize_single_token(token: str) -> str:
    if X_RANGE_PATTERN.match(token):
        lower, upper = _x_range_to_bounds(token)
        return f">={lower} <{upper}"
    return token


def _resolve_range_with_operators(range_token: str, operators: list[str]) -> str:
    lower, upper = _range_token_to_bounds(range_token)
    lower_ops = [t for t in operators if t.startswith(">")]
    upper_ops = [t for t in operators if t.startswith("<")]
    result: list[str] = []
    result.extend(lower_ops if lower_ops else [f">={lower}"])
    result.extend(upper_ops if upper_ops else [f"<{upper}"])
    return " ".join(result)


def _normalize_npm_range_part(part: str) -> str:
    hyphen_match = HYPHEN_RANGE_PATTERN.match(part)
    if hyphen_match:
        return _convert_hyphen_range(hyphen_match.group(1), hyphen_match.group(2))

    normalized = re.sub(r"([><=!~^]+)\s+", r"\1", part)
    tokens = normalized.split()

    if len(tokens) <= 1:
        return _normalize_single_token(tokens[0]) if tokens else normalized

    range_tokens = [t for t in tokens if _is_range_token(t)]
    caret_tilde = [t for t in tokens if t[0] in "~^"]
    operators = [t for t in tokens if t[0] in "><=!"]

    if caret_tilde and operators:
        return _resolve_caret_tilde_range(caret_tilde[0], operators)

    if caret_tilde or any(t.startswith("=") for t in operators):
        return " ".join(caret_tilde) if caret_tilde else " ".join(operators)

    if range_tokens and operators:
        return _resolve_range_with_operators(range_tokens[0], operators)

    non_range = [t for t in tokens if not _is_range_token(t)]
    return " ".join(non_range) if non_range else normalized


def _normalize_version_or_list(version_str: str) -> str:
    parts = [p.strip() for p in version_str.split("||")]
    normalized: list[str] = []
    for p in parts:
        stripped = _normalize_npm_range_part(p)
        if re.fullmatch(r"\d+(?:\.\d+)?", stripped):
            normalized.append(f"{stripped}.*")
        else:
            normalized.append(stripped)
    return " || ".join(normalized)


def normalize_npm_alias(version_or_alias: str) -> tuple[str | None, str]:
    if version_or_alias.startswith("npm:"):
        parts = version_or_alias.removeprefix("npm:").rsplit("@", 1)
        if len(parts) == 2:
            return parts[0], parts[1]
    return None, version_or_alias


def read_package_json(location: Location) -> IndexedDict[str, ParsedValue] | None:
    if not (location.coordinates and location.coordinates.real_path):
        return None

    lockfile_path = Path(location.coordinates.real_path)
    package_json_path = lockfile_path.parent / "package.json"

    if not package_json_path.exists():
        return None

    try:
        with package_json_path.open("r", encoding="utf-8") as f:
            content = parse_json_with_tree_sitter(f.read())
            if isinstance(content, IndexedDict):
                return content
            return None
    except Exception as exc:  # noqa: BLE001
        LOGGER.exception("Error reading package.json", extra={"extra": {"error": exc}})
        return None


def get_direct_dependencies_from_package_json(location: Location) -> set[str]:
    pkg_json = read_package_json(location)
    if not pkg_json:
        return set()

    direct: set[str] = set()
    deps = pkg_json.get("dependencies")
    dev_deps = pkg_json.get("devDependencies")
    opt_deps = pkg_json.get("optionalDependencies")

    if isinstance(deps, IndexedDict):
        direct.update(deps.keys())
    if isinstance(dev_deps, IndexedDict):
        direct.update(dev_deps.keys())
    if isinstance(opt_deps, IndexedDict):
        direct.update(opt_deps.keys())

    return direct


def get_dev_dependencies_from_package_json(location: Location) -> set[str]:
    pkg_json = read_package_json(location)
    if not pkg_json:
        return set()

    dev_deps = pkg_json.get("devDependencies")
    if isinstance(dev_deps, IndexedDict):
        return set(dev_deps.keys())

    return set()


def get_direct_dependencies_with_fallback(location: Location, fallback: list[str]) -> list[str]:
    pkg_json = read_package_json(location)
    if not pkg_json:
        return fallback

    direct: set[str] = set()
    deps = pkg_json.get("dependencies")
    dev_deps = pkg_json.get("devDependencies")
    opt_deps = pkg_json.get("optionalDependencies")

    if isinstance(deps, IndexedDict):
        direct.update(deps.keys())
    if isinstance(dev_deps, IndexedDict):
        direct.update(dev_deps.keys())
    if isinstance(opt_deps, IndexedDict):
        direct.update(opt_deps.keys())

    return list(direct)


def create_root_package_from_package_json(
    location: Location,
    package_manager: PackageManager = PackageManager.NPM,
) -> Package | None:
    pkg_json = read_package_json(location)
    if not pkg_json:
        return None

    root_name = pkg_json.get("name")
    root_version = pkg_json.get("version")

    if not isinstance(root_name, str) or not isinstance(root_version, str):
        return None

    root_location = get_enriched_location(
        location,
        line=1,
        is_transitive=False,
        is_dev=False,
        package_manager=package_manager,
    )
    update_root: dict[str, DependencyType] = {"dependency_type": DependencyType.ROOT}
    root_location = root_location.model_copy(deep=True, update=update_root)

    return new_simple_npm_package(root_location, root_name, root_version)


def find_root_location(root_package: Package, lockfile_path: str) -> Location | None:
    for loc in root_package.locations:
        if loc.coordinates and loc.coordinates.real_path == lockfile_path:
            return loc
    return None


def group_packages_by_name(
    packages: list[Package], root_package_id: str
) -> dict[str, list[Package]]:
    packages_by_name: dict[str, list[Package]] = {}
    for pkg in packages:
        if pkg.id_ != root_package_id:
            packages_by_name.setdefault(pkg.name, []).append(pkg)
    return packages_by_name


def add_dependency_to_root_relationship(
    dep_pkg: Package,
    root_package: Package,
    root_location: Location,
    lockfile_path: str,
    relationships: list[Relationship],
) -> None:
    for dep_loc in dep_pkg.locations:
        if dep_loc.coordinates and dep_loc.coordinates.real_path == lockfile_path:
            from_id = f"{dep_pkg.id_}@{dep_loc.location_id()}"
            to_id = f"{root_package.id_}@{root_location.location_id()}"
            relationships.append(
                Relationship(
                    from_=from_id,
                    to_=to_id,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
            )
            break


def is_version_compatible(package_version: str, required_version: str) -> bool:
    version_str = required_version.strip()
    if version_str in ("*", "*.*", "*.*.*"):
        return True

    version_str = _normalize_version_or_list(version_str)

    return match_version_ranges(package_version, version_str)


def _is_exact_version(version_str: str) -> bool:
    if X_RANGE_PATTERN.match(version_str):
        return False
    return not any(char in version_str for char in ("^", "~", "*", ">", "<", "=", " "))


def build_relationships(
    packages: list[Package],
    dependency_map: DependencyMap,
    location: Location,
    key_map: dict[tuple[str, int], str] | None = None,
) -> list[Relationship]:
    """Build relationships between packages.

    Args:
        packages: List of packages to create relationships for
        dependency_map: Map of package names to their dependencies
        location: Location of the lockfile
        key_map: Optional map for yarn lockfiles

    Returns:
        List of relationships

    """
    lockfile_path = location.path()
    relationships: list[Relationship] = []

    packages_by_name: dict[str, list[Package]] = {}
    for pkg in packages:
        packages_by_name.setdefault(pkg.name, []).append(pkg)

    for package in packages:
        package_lock_locations = [loc for loc in package.locations if loc.path() == lockfile_path]
        if not package_lock_locations:
            continue

        for package_location in package_lock_locations:
            if key_map is not None:
                # Use key_map to get dependencies (for yarn and pnpm)
                relationships.extend(
                    _create_relationships_with_key_map(
                        package,
                        package_location,
                        key_map,
                        dependency_map,
                        packages_by_name,
                        lockfile_path,
                    )
                )
            else:
                # Fallback for simple name lookup (package-lock v1/v2)
                dependencies = dependency_map.get(package.name)
                if not dependencies:
                    continue

                relationships.extend(
                    _create_simple_relationships(
                        package, package_location, dependencies, packages_by_name, lockfile_path
                    )
                )

    return relationships


def _create_relationships_with_key_map(  # noqa: PLR0913
    package: Package,
    package_location: Location,
    key_map: dict[tuple[str, int], str],
    dependency_map: DependencyMap,
    packages_by_name: dict[str, list[Package]],
    lockfile_path: str,
) -> list[Relationship]:
    line = package_location.coordinates.line if package_location.coordinates else None
    if line is None:
        return []

    dep_key = key_map.get((package.name, line))
    if not dep_key:
        return []

    dependencies = dependency_map.get(dep_key)
    if not dependencies:
        return []

    return _create_simple_relationships(
        package, package_location, dependencies, packages_by_name, lockfile_path
    )


def _create_simple_relationships(
    package: Package,
    package_location: Location,
    dependencies: DependencyRequirements,
    packages_by_name: dict[str, list[Package]],
    lockfile_path: str,
) -> list[Relationship]:
    relationships: list[Relationship] = []

    for dependency_name, required_version in dependencies.items():
        use_exact = _is_exact_version(required_version)

        for dep_pkg in packages_by_name.get(dependency_name, []):
            if use_exact:
                if dep_pkg.version != required_version:
                    continue
            elif not is_version_compatible(dep_pkg.version, required_version):
                continue

            dep_locations = [loc for loc in dep_pkg.locations if loc.path() == lockfile_path]

            for dep_location in dep_locations:
                from_id = f"{dep_pkg.id_}@{dep_location.location_id()}"
                to_id = f"{package.id_}@{package_location.location_id()}"
                relationships.append(
                    Relationship(
                        from_=from_id,
                        to_=to_id,
                        type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                    )
                )

    return relationships



import logging
from zipfile import ZipInfo

from labels.model.ecosystem_data.java import JavaManifest
from labels.model.file import Location
from labels.utils.zip import contents_from_zip, new_zip_glob_match

LOGGER = logging.getLogger(__name__)


class ManifestParser:
    def __init__(
        self, *, file_manifest: list[ZipInfo], archive_path: str, location: Location
    ) -> None:
        self.file_manifest = file_manifest
        self.archive_path = archive_path
        self.location = location

    def parse(self) -> JavaManifest | None:
        manifest_matches = self._find_manifest_files()
        if not self._validate_matches(manifest_matches):
            return None

        contents = self._read_manifest(manifest_matches[0])
        if not contents:
            return None

        manifest = self._parse_java_manifest(contents)

        if self._should_exclude(manifest):
            return None

        return manifest

    def _find_manifest_files(self) -> list[str]:
        return new_zip_glob_match(
            self.file_manifest, ("*META-INF/MANIFEST.MF",), case_sensitive=True
        )

    def _validate_matches(self, matches: list[str]) -> bool:
        if len(matches) != 1:
            if len(matches) > 1:
                LOGGER.debug("Found multiple manifests in the jar: %s", matches)
            else:
                LOGGER.debug("No manifests found in the jar")
            return False
        return True

    def _read_manifest(self, manifest_path: str) -> str | None:
        contents = contents_from_zip(self.archive_path, manifest_path)
        return contents.get(manifest_path) if contents else None

    def _should_exclude(self, manifest: JavaManifest | None) -> bool:
        if not manifest:
            LOGGER.debug("Failed to parse java manifest: %s", self.location)
            return True

        if manifest.main and "Weave-Classes" in manifest.main:
            LOGGER.debug("Excluding archive due to Weave-Classes: %s", self.location)
            return True

        return False

    def _parse_java_manifest(self, content: str) -> JavaManifest | None:
        sections = self._process_section(content)

        return self._build_manifest(sections)

    def _process_section(self, content: str) -> list[dict[str, str]]:
        sections: list[dict[str, str]] = []
        current_section = -1
        last_key = ""

        for line in content.splitlines():
            if not line.strip():
                last_key = ""
                continue

            last_key = self._parse_line(line, last_key, current_section, sections)
            current_section = len(sections) - 1

        return sections

    def _parse_line(
        self, line: str, last_key: str, current_section: int, sections: list[dict[str, str]]
    ) -> str:
        line = line.strip()

        idx = line.find(":")
        if idx == -1:
            return last_key

        key = line[:idx].strip()
        value = line[idx + 1 :].strip()

        if not key:
            return last_key

        if last_key == "" or current_section == -1:
            sections.append({})
            current_section += 1

        sections[current_section][key] = value

        return key

    def _build_manifest(self, sections: list[dict[str, str]]) -> JavaManifest | None:
        if sections:
            main_section = sections[0]
            other_sections = sections[1:] if len(sections) > 1 else None
            return JavaManifest(main=main_section, sections=other_sections)

        return None



from pathlib import Path
from typing import NamedTuple, TextIO

from bs4 import BeautifulSoup, Tag

from labels.model.ecosystem_data.java import JavaArchive, JavaPomProperties
from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.java.utils.model import PomContext
from labels.parsers.cataloger.java.utils.package_builder import JavaPackageSpec, new_java_package
from labels.parsers.cataloger.java.utils.parsing_utils import get_direct_child_text
from labels.parsers.cataloger.java.utils.version_resolver import resolve_version
from labels.parsers.cataloger.utils import get_enriched_location


class PomEvaluationContext(NamedTuple):
    resolver: Resolver
    parent_info: dict[str, str]
    current_project: Tag
    current_pom_path: str


def parse_pom_xml(
    resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    pom_tree = _safe_parse_pom_xml(reader)
    if not pom_tree:
        return [], []

    pom_context = _build_pom_context(pom_tree, reader, resolver)
    if not pom_context:
        return [], []

    packages = _collect_packages(pom_context, reader.location)

    return packages, []


def _safe_parse_pom_xml(reader: LocationReadCloser) -> BeautifulSoup | None:
    try:
        return BeautifulSoup(reader.read_closer, features="html.parser")
    except (AssertionError, UnicodeError):
        return None


def _build_pom_context(
    pom_tree: BeautifulSoup,
    reader: LocationReadCloser,
    resolver: Resolver | None,
) -> PomContext | None:
    valid_project_and_dependencies = _get_valid_project_and_dependencies(pom_tree)
    if not valid_project_and_dependencies:
        return None

    project, dependencies = valid_project_and_dependencies

    parent_info = _get_parent_info(project)

    if resolver and parent_info:
        evaluation_context = PomEvaluationContext(
            resolver=resolver,
            parent_info=parent_info,
            current_project=project,
            current_pom_path=str(reader.location.access_path),
        )
        parent_version_properties, manage_deps = _evaluate_pom_files_in_project(evaluation_context)
    else:
        parent_version_properties = _get_properties(pom_tree)
        manage_deps = _get_deps_management(pom_tree)

    return PomContext(
        project=project,
        dependencies=dependencies,
        parent_info=parent_info,
        parent_version_properties=parent_version_properties,
        manage_deps=manage_deps,
    )


def _get_valid_project_and_dependencies(pom_tree: BeautifulSoup) -> tuple[Tag, Tag] | None:
    project = pom_tree.find("project")
    if not isinstance(project, Tag):
        return None

    if str(project.get("xmlns")) != "http://maven.apache.org/POM/4.0.0":
        return None

    dependencies = project.find("dependencies", recursive=False)
    if not isinstance(dependencies, Tag):
        return None

    return project, dependencies


def _get_parent_info(project: Tag) -> dict[str, str] | None:
    parent = project.find_next("parent")
    if not isinstance(parent, Tag):
        return None

    group_node = get_direct_child_text(parent, "groupid")
    artifact_node = get_direct_child_text(parent, "artifactid")
    version_node = get_direct_child_text(parent, "version")

    if not (group_node and artifact_node and version_node):
        return None

    return {
        "group": group_node,
        "artifact": artifact_node,
        "version": version_node,
    }


def _evaluate_pom_files_in_project(
    context: PomEvaluationContext,
) -> tuple[dict[str, str], dict[str, str]]:
    parent_pom = _find_parent_pom(context)
    properties_vars = _get_properties(parent_pom) if parent_pom else {}
    manage_deps = _get_deps_management(parent_pom) if parent_pom else {}

    manage_deps.update(_get_deps_management(context.current_project))

    return properties_vars, manage_deps


def _find_parent_pom(context: PomEvaluationContext) -> Tag | None:
    for loc in context.resolver.files_by_glob("**/pom.xml", "pom.xml"):
        content = context.resolver.file_contents_by_location(loc)
        if not content:
            continue

        pom_project = _get_pom_project(content)
        if not pom_project:
            continue

        if _is_parent_pom(pom_project, context.parent_info) and _is_module_parent(
            pom_project, context.current_pom_path
        ):
            return pom_project

    return None


def _get_pom_project(content: TextIO) -> Tag | None:
    file_content = BeautifulSoup(content, features="html.parser")
    pom_project = file_content.find("project")
    if not isinstance(pom_project, Tag):
        return None

    if str(pom_project.get("xmlns")) == "http://maven.apache.org/POM/4.0.0":
        return pom_project

    return None


def _is_parent_pom(root_pom_project: Tag, parent_info: dict[str, str]) -> bool:
    group = root_pom_project.find("groupid", recursive=False)
    artifact = root_pom_project.find("artifactid", recursive=False)
    version = root_pom_project.find("version", recursive=False)

    if not (group and artifact and version):
        return False

    return (
        group.get_text() == parent_info["group"]
        and artifact.get_text() == parent_info["artifact"]
        and version.get_text() == parent_info["version"]
    )


def _is_module_parent(parent_pom: Tag, pom_module: str) -> bool:
    base_module_name = Path(pom_module).parent.name

    modules_sections: list[Tag] = parent_pom.find_all("modules")
    for modules_section in modules_sections:
        module_tags: list[Tag] = modules_section.find_all("module")
        for module_tag in module_tags:
            if module_tag.get_text() == base_module_name:
                return True

    return False


def _get_properties(project_pom: Tag) -> dict[str, str]:
    properties_dict: dict[str, str] = {}

    properties_sections: list[Tag] = project_pom.find_all("properties", limit=2)
    for properties_section in properties_sections:
        for prop in properties_section.children:
            if isinstance(prop, Tag) and prop.name:
                properties_dict[prop.name.lower()] = prop.get_text()

    return properties_dict


def _get_deps_management(pom_tree: Tag) -> dict[str, str]:
    deps_info: dict[str, str] = {}

    management_sections: list[Tag] = pom_tree.find_all("dependencymanagement")
    for manage in management_sections:
        dep_tags: list[Tag] = manage.find_all("dependency", recursive=True)
        for dependency in dep_tags:
            groupid_node = dependency.find("groupid")
            artifactid_node = dependency.find("artifactid")
            version_node = dependency.find("version")

            if not (groupid_node and artifactid_node and version_node):
                continue

            group = groupid_node.get_text()
            artifact = artifactid_node.get_text()
            version = version_node.get_text()
            key = f"{group}:{artifact}"
            deps_info[key] = version

    return deps_info


def _collect_packages(pom_context: PomContext, location: Location) -> list[Package]:
    packages: list[Package] = []

    dep_tags: list[Tag] = pom_context.dependencies.find_all("dependency")
    for dependency in dep_tags:
        package_spec = _build_package_spec(pom_context, dependency, location)
        if not package_spec:
            continue

        package = new_java_package(package_spec)
        if package:  # pragma: no branch
            packages.append(package)

    return packages


def _build_package_spec(
    pom_context: PomContext, dependency: Tag, location: Location
) -> JavaPackageSpec | None:
    name = get_direct_child_text(dependency, "artifactid")
    if not name:
        return None

    group_id = get_direct_child_text(dependency, "groupid")
    full_name = f"{group_id}:{name}" if group_id else name

    pom_properties = JavaPomProperties(
        group_id=group_id,
        artifact_id=name,
        version=get_direct_child_text(dependency, "version"),
    )

    version = resolve_version(pom_context, pom_properties, full_name)
    if not _is_valid_version(version):
        return None

    scope = get_direct_child_text(dependency, "scope")
    is_dev = scope in ["test", "provided"]

    new_location = get_enriched_location(
        location,
        line=dependency.version.sourceline if dependency.version else dependency.sourceline,
        is_transitive=False,
        is_dev=is_dev,
        package_manager=PackageManager.MAVEN,
    )

    return JavaPackageSpec(
        composed_name=full_name,
        simple_name=name,
        version=str(version),
        location=new_location,
        ecosystem_data=JavaArchive(pom_properties=pom_properties),
    )


def _is_valid_version(version: str | None) -> bool:
    return version is not None and not version.startswith("${")



from collections.abc import Callable
from typing import cast

import reactivex
from reactivex import (
    Observable,
)
from reactivex import (
    operators as ops,
)
from reactivex.scheduler import (
    ThreadPoolScheduler,
)

from labels.model.parser import Request
from labels.parsers.cataloger.dart.cataloger import (
    on_next_dart,
)
from labels.parsers.cataloger.dotnet.cataloger import (
    on_next_dotnet,
)
from labels.parsers.cataloger.golang.cataloger import (
    on_next_golang,
)
from labels.parsers.cataloger.java.cataloger import (
    on_next_java,
)
from labels.parsers.cataloger.javascript.cataloger import (
    on_next_javascript,
)
from labels.parsers.cataloger.php.cataloger import (
    on_next_php,
)
from labels.parsers.cataloger.python.cataloger import (
    on_next_python,
)
from labels.parsers.cataloger.ruby.cataloger import (
    on_next_ruby,
)
from labels.parsers.cataloger.swift.cataloger import (
    on_next_swift,
)


def handle_parser(
    scheduler: ThreadPoolScheduler,
) -> Callable[[Observable[str]], Observable[Request]]:
    def _apply_parsers(source: Observable[str]) -> Observable[Request]:
        def fanout(item: str) -> Observable[Request]:
            return cast(
                "Observable[Request]",
                reactivex.merge(
                    (on_next_python(reactivex.just(item, scheduler))),
                    (on_next_java(reactivex.just(item, scheduler))),
                    (on_next_javascript(reactivex.just(item, scheduler))),
                    (on_next_dotnet(reactivex.just(item, scheduler))),
                    (on_next_ruby(reactivex.just(item, scheduler))),
                    (on_next_php(reactivex.just(item, scheduler))),
                    (on_next_swift(reactivex.just(item, scheduler))),
                    (on_next_dart(reactivex.just(item, scheduler))),
                    (on_next_golang(reactivex.just(item, scheduler))),
                ),
            )

        return source.pipe(
            ops.flat_map(fanout),
        )

    return _apply_parsers



import functools
import hashlib
import io
import logging
import os
import pickle  # nosec
import socket
from collections.abc import Callable
from contextlib import suppress
from typing import ParamSpec, TypeVar, cast

import diskcache
import diskcache.core
from diskcache import Lock
from platformdirs import user_cache_dir
from redis import Redis

LOGGER = logging.getLogger(__name__)
REDIS_CACHE_ENDPOINT = os.environ.get("CACHE_URL")
REDIS_CLIENT: Redis | None = None

P = ParamSpec("P")
R = TypeVar("R")
TFun = Callable[P, R]

if REDIS_CACHE_ENDPOINT:
    try:
        host, _port = REDIS_CACHE_ENDPOINT.split(":", maxsplit=1)
        port = int(_port)

        # Try to connect to redis endpoint
        socket.create_connection((host, port), timeout=2)

        REDIS_CLIENT = cast(
            "Redis",
            Redis(
                host=host,
                port=port,
                ssl=True,
                username="batch-rw-user",
                password=os.environ.get("CACHE_BATCH_RW_USER_PASSWORD"),
            ),
        )
    except (OSError, ValueError):
        REDIS_CLIENT = None

DISK_CACHE = diskcache.Cache(user_cache_dir("fluid-labels", "fluidattacks"))


# 2 weeks
TTL = 604800 * 2

# Whitelist of allowed modules and classes for pickling/unpickling
_PICKLE_ALLOWED_CLASSES: frozenset[tuple[str, str]] = frozenset(
    {
        # Built-in types
        ("builtins", "dict"),
        ("builtins", "list"),
        ("builtins", "set"),
        ("builtins", "frozenset"),
        ("builtins", "tuple"),
        ("builtins", "bytes"),
        ("builtins", "bytearray"),
        # Collections
        ("collections", "OrderedDict"),
        ("collections", "defaultdict"),
        # datetime types
        ("datetime", "datetime"),
        ("datetime", "date"),
        ("datetime", "time"),
        ("datetime", "timedelta"),
    }
)


class RestrictedUnpickler(pickle.Unpickler):  # noqa: S301 - mitigated via _PICKLE_ALLOWED_CLASSES allowlist
    """Unpickler that only allows specific safe classes."""

    def find_class(self, module: str, name: str) -> type:
        if (module, name) in _PICKLE_ALLOWED_CLASSES:
            return cast("type", super().find_class(module, name))

        err_msg = f"Forbidden class in cache: {module}.{name}"
        raise pickle.UnpicklingError(err_msg)

    def load(self) -> object:
        return cast("object", super().load())


def _pickle_loads(data: object) -> object:
    """Deserialize bytes data using RestrictedUnpickler.

    Returns deserialized object if safe, otherwise raises exception
    """
    if not isinstance(data, bytes):
        raise pickle.UnpicklingError

    return RestrictedUnpickler(io.BytesIO(data)).load()


def _pickle_dumps(item: object) -> bytes:
    """Serialize and validate an object using RestrictedUnpickler.

    Returns serialized bytes if safe, otherwise raises exception
    """
    data = pickle.dumps(item)
    RestrictedUnpickler(io.BytesIO(data)).load()
    return data


def make_hashable(item: object) -> str:
    serialized_object = pickle.dumps(item)

    return hashlib.sha256(serialized_object).hexdigest()


def safe_generate_cache_key(
    func: Callable[P, R],
    args: object,
    kwargs: object,
) -> str | None:
    try:
        key = f"{func.__module__}.{func.__name__}-"
        key += str(make_hashable((args, kwargs)))
        return hashlib.sha256(key.encode()).hexdigest()
    except Exception:
        LOGGER.exception("Failed to generate cache key for function: %s", func.__name__)
        return None


def safe_save_to_cache(result: object, cache_key: str, function_name: str) -> None:
    try:
        data = _pickle_dumps(result)
        with suppress(diskcache.core.Timeout), Lock(DISK_CACHE, "dual_cache"):
            DISK_CACHE.set(cache_key, data, expire=TTL, retry=True)  # Save with TTL
        if REDIS_CLIENT:
            REDIS_CLIENT.setex(cache_key, TTL, data)  # Save with TTL
    except Exception:
        LOGGER.exception("Failed to pickle result for function: %s", function_name)


_SERIALIZATION_ERROR = object()


def safe_retrieve_from_cache(cache_key: str) -> object:
    with suppress(Exception):
        # Try to recover from disk cache
        if cached_data := DISK_CACHE.get(cache_key):
            return _pickle_loads(cached_data)

        # Try to recover from Redis cache (ElastiCache) if available
        if REDIS_CLIENT and (cached_data := REDIS_CLIENT.get(cache_key)):
            result = _pickle_loads(cached_data)
            # Save to disk cache
            with Lock(DISK_CACHE, "dual_cache"):
                DISK_CACHE.set(cache_key, cached_data, expire=TTL, retry=True)
            return result
    # Anything other than a cache hit returns a serialization error
    return _SERIALIZATION_ERROR


def dual_cache(func: Callable[P, R]) -> Callable[P, R]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        if os.environ.get("SKIP_FA_LABELS_CACHE_USE"):
            return func(*args, **kwargs)

        cache_key = safe_generate_cache_key(func, args, kwargs)
        if cache_key:
            cached_result = safe_retrieve_from_cache(cache_key)
            # None or falsy values might be stored on the cache, so this check is necessary
            if cached_result is not _SERIALIZATION_ERROR:
                return cast("R", cached_result)

        # Run the function and store the result in both caches if serializable
        result = func(*args, **kwargs)
        if cache_key:
            safe_save_to_cache(result, cache_key, func.__name__)
        return result

    return wrapper



from collections.abc import Iterable

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.metadata import Artifact, Digest, HealthMetadata
from labels.model.package import Ecosystem, Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection import toml
from labels.utils.strings import normalize_name


def parse_poetry_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = toml.parse_toml_with_tree_sitter(reader.read_closer.read())

    packages = _collect_packages(file_content, reader)
    relationships = _collect_relationships(file_content, packages)

    return packages, relationships


def _select_hash(files: IndexedList[ParsedValue]) -> str | None:
    selected_hash: str | None = None
    for entry in files:
        if not isinstance(entry, IndexedDict):
            continue
        filename = str(entry.get("file", ""))
        hash_value = str(entry.get("hash", ""))
        if not hash_value:
            continue
        if filename.endswith(".tar.gz"):
            return hash_value
        if selected_hash is None and filename.endswith(".whl"):
            selected_hash = hash_value
    return selected_hash


def _extract_integrity(raw_package: IndexedDict[str, ParsedValue]) -> HealthMetadata | None:
    files = raw_package.get("files")
    if not isinstance(files, IndexedList):
        return None

    selected_hash = _select_hash(files)
    if not selected_hash or ":" not in selected_hash:
        return None

    algorithm, digest_value = selected_hash.split(":", 1)
    if not algorithm or not digest_value:
        return None

    return HealthMetadata(
        artifact=Artifact(
            integrity=Digest(algorithm=algorithm, value=digest_value),
        ),
    )


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue],
    reader: LocationReadCloser,
) -> list[Package]:
    packages = []
    toml_pkgs = file_content.get("package")
    if not isinstance(toml_pkgs, IndexedList):
        return []

    for raw_package in toml_pkgs:
        if not isinstance(raw_package, IndexedDict):
            continue
        name = str(raw_package.get("name", ""))
        version = str(raw_package.get("version", ""))

        if not version:
            continue

        new_location = (
            get_enriched_location(
                reader.location,
                line=raw_package.get_key_position("version").start.line,
                package_manager=PackageManager.POETRY,
            )
            if isinstance(raw_package, IndexedDict)
            else reader.location
        )

        package = new_python_package(name=name, version=version, location=new_location)
        if package:
            package.health_metadata = _extract_integrity(raw_package)
            packages.append(package)

    return packages


def _get_dependencies(
    package: ParsedValue,
    packages: list[Package],
) -> tuple[Package | None, IndexedDict[str, ParsedValue]] | None:
    if not isinstance(package, IndexedDict):
        return None

    package_name = package.get("name")
    if not isinstance(package_name, str):
        return None

    _pkg = _find_package_by_name(packages, package_name)

    deps = package.get("dependencies")
    if not isinstance(deps, IndexedDict):
        return None

    return _pkg, deps


def _find_package_by_name(packages: list[Package], name: str) -> Package | None:
    return next((p for p in packages if p.name == normalize_name(name, Ecosystem.PYPI)), None)


def _collect_relationships(
    toml_content: IndexedDict[str, ParsedValue],
    packages: list[Package],
) -> list[Relationship]:
    relationships = []
    toml_pkgs = toml_content.get("package")

    if not isinstance(toml_pkgs, IndexedList):
        return []

    for package in toml_pkgs:
        pkg_with_deps = _get_dependencies(package, packages)

        if not pkg_with_deps:
            continue

        pkg, deps = pkg_with_deps

        if not pkg or not deps:
            continue

        relationships.extend(_create_relationships_for_package(pkg, deps.keys(), packages))

    return relationships


def _create_relationships_for_package(
    pkg: Package, dependency_names: Iterable[str], packages: list[Package]
) -> list[Relationship]:
    relationships = []
    for dep_name in dependency_names:
        dep_pkg = _find_package_by_name(packages, dep_name)
        if dep_pkg:
            relationships.append(
                Relationship(
                    from_=dep_pkg.id_,
                    to_=pkg.id_,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
            )
    return relationships



from enum import Enum
from typing import Any

from pydantic import BaseModel

AdvisoryRecord = tuple[
    str,  # id
    str,  # source
    str,  # vulnerable_version
    str | None,  # severity_level
    str | None,  # severity_v4
    str | None,  # epss
    str | None,  # details
    str | None,  # percentile
    str | None,  # cwe_ids
    str | None,  # cve_finding
    int,  # auto_approve
    str | None,  # fixed_versions
    int,  # kev_catalog
    str | None,  # platform_version
]


class UpgradeType(str, Enum):
    UNKNOWN = "unknown"
    MAJOR = "major"
    MINOR = "minor"
    PATCH = "patch"


class SeverityCVES(BaseModel):
    critical: list[str] | None = None
    high: list[str] | None = None
    medium: list[str] | None = None
    low: list[str] | None = None
    total: int


class ResidualCVES(BaseModel):
    remedied: SeverityCVES | None = None
    maintained: SeverityCVES | None = None
    introduced: SeverityCVES | None = None


class FixMetadata(BaseModel):
    fix_version: str
    upgrade_type: UpgradeType
    breaking_change: bool
    residual_cves: ResidualCVES | None = None


class AdvisoryFixMetadataNew(BaseModel):
    closest_min_fix: FixMetadata
    closest_safe_fix: FixMetadata | None = None
    closest_complete_fix: FixMetadata | None = None


class AdvisoryFixMetadata(BaseModel):
    closest_fix_version: str
    upgrade_type: UpgradeType
    breaking_change: bool
    closest_safe_version: str | None = None


class Advisory(BaseModel):
    id: str
    vulnerable_version: str
    source: str
    cpes: list[str]
    severity_level: str = "Low"
    ecosystem: str
    platform_version: str | None = None
    fixed_versions: list[str] | None = None
    fix_metadata: AdvisoryFixMetadata | None = None
    fix_metadata_new: AdvisoryFixMetadataNew | None = None
    details: str | None = None
    epss: float = 0.0
    percentile: float = 0.0
    severity_v4: str | None = None
    cwe_ids: list[str] | None = None
    cve_finding: str | None = None
    auto_approve: bool = False
    upstream_package: str | None = None
    kev_catalog: bool = False

    def serialize(self) -> dict[str, Any]:  # type: ignore[explicit-any]
        return {
            "cpes": self.cpes,
            "description": self.details,
            "epss": self.epss,
            "id": self.id,
            "percentile": self.percentile,
            "severity": self.severity_level,
            "urls": [self.source],  # type: ignore[misc]
            "version_constraint": self.vulnerable_version,
            "platform_version": self.platform_version,
            "fixed_versions": self.fixed_versions,
            "fix_metadata": self.fix_metadata.model_dump(exclude_none=True)  # type: ignore[misc]
            if self.fix_metadata
            else None,
            "fix_metadata_new": self.fix_metadata_new.model_dump(exclude_none=True)  # type: ignore[misc]
            if self.fix_metadata_new
            else None,
            "cvss4": self.severity_v4,
            "cwe_ids": self.cwe_ids,
            "cve_finding": self.cve_finding,
            "auto_approve": self.auto_approve,
            "upstream_package": self.upstream_package,
            "kev_catalog": self.kev_catalog,
        }



FLUID_SBOM_JSON_SCHEMA = {
    "type": "object",
    "additionalProperties": False,
    "properties": {
        "sbom_details": {
            "type": "object",
            "additionalProperties": False,
            "properties": {
                "name": {"type": "string"},
                "timestamp": {"type": "string", "format": "date-time"},
                "tool": {"type": "string"},
                "organization": {"type": "string"},
                "version": {"type": ["string", "null"]},
            },
            "required": [
                "name",
                "version",
                "timestamp",
                "tool",
                "organization",
            ],
        },
        "packages": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "id": {"type": "string"},
                    "name": {"type": "string"},
                    "version": {"type": "string"},
                    "locations": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "path": {"type": "string"},
                                "line": {"type": ["integer", "null"]},
                                "layer": {"type": ["string", "null"]},
                                "dependency_type": {"type": "string"},
                                "scope": {"type": "string"},
                                "reachable_cves": {"type": "array", "items": {"type": ["string"]}},
                                "location_id": {"type": "string"},
                                "top_parents": {
                                    "type": ["array", "null"],
                                    "items": {"type": "string"},
                                },
                                "dependency_chains": {
                                    "type": ["array", "null"],
                                    "items": {"type": "object"},
                                },
                                "package_manager": {"type": "string"},
                            },
                            "required": ["path", "line", "layer", "package_manager"],
                        },
                    },
                    "language": {"type": "string"},
                    "found_by": {"type": "string"},
                    "package_url": {"type": "string"},
                    "ecosystem": {"type": "string"},
                    "health_metadata": {
                        "type": ["object", "null"],
                        "additionalProperties": False,
                        "properties": {
                            "artifact": {
                                "type": ["object", "null"],
                                "properties": {
                                    "integrity": {
                                        "type": ["object", "null"],
                                        "additionalProperties": False,
                                        "properties": {
                                            "algorithm": {"type": ["string", "null"]},
                                            "value": {"type": ["string", "null"]},
                                        },
                                        "required": ["algorithm", "value"],
                                    },
                                },
                            },
                        },
                    },
                    "safe_versions": {
                        "type": ["array", "null"],
                        "items": {"type": "string"},
                    },
                    "advisories": {
                        "type": "array",
                        "items": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "cpes": {
                                    "type": "array",
                                    "items": {"type": "string"},
                                },
                                "description": {"type": ["string", "null"]},
                                "epss": {"type": "number"},
                                "id": {"type": "string"},
                                "percentile": {"type": "number"},
                                "severity": {"type": "string"},
                                "urls": {
                                    "type": "array",
                                    "items": {"type": "string"},
                                },
                                "version_constraint": {"type": "string"},
                                "platform_version": {"type": ["string", "null"]},
                                "cvss4": {"type": ["string", "null"]},
                                "cwe_ids": {
                                    "type": ["array", "null"],
                                    "items": {"type": "string"},
                                },
                                "fixed_versions": {
                                    "type": ["array", "null"],
                                    "items": {"type": "string"},
                                },
                                "fix_metadata_new": {
                                    "type": ["object", "null"],
                                    "additionalProperties": False,
                                    "properties": {
                                        "closest_min_fix": {
                                            "type": ["object"],
                                            "additionalProperties": False,
                                            "properties": {
                                                "fix_version": {"type": "string"},
                                                "upgrade_type": {
                                                    "type": "string",
                                                    "enum": ["unknown", "major", "minor", "patch"],
                                                },
                                                "breaking_change": {"type": "boolean"},
                                                "residual_cves": {
                                                    "type": ["object", "null"],
                                                    "additionalProperties": False,
                                                    "properties": {
                                                        "remedied": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                        "maintained": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                        "introduced": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                    },
                                                },
                                            },
                                            "required": [
                                                "fix_version",
                                                "upgrade_type",
                                                "breaking_change",
                                            ],
                                        },
                                        "closest_safe_fix": {
                                            "type": ["object", "null"],
                                            "additionalProperties": False,
                                            "properties": {
                                                "fix_version": {"type": "string"},
                                                "upgrade_type": {
                                                    "type": "string",
                                                    "enum": ["unknown", "major", "minor", "patch"],
                                                },
                                                "breaking_change": {"type": "boolean"},
                                                "residual_cves": {
                                                    "type": ["object", "null"],
                                                    "additionalProperties": False,
                                                    "properties": {
                                                        "remedied": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                        "maintained": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                        "introduced": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                    },
                                                },
                                            },
                                            "required": [
                                                "fix_version",
                                                "upgrade_type",
                                                "breaking_change",
                                            ],
                                        },
                                        "closest_complete_fix": {
                                            "type": ["object", "null"],
                                            "additionalProperties": False,
                                            "properties": {
                                                "fix_version": {"type": "string"},
                                                "upgrade_type": {
                                                    "type": "string",
                                                    "enum": ["unknown", "major", "minor", "patch"],
                                                },
                                                "breaking_change": {"type": "boolean"},
                                                "residual_cves": {
                                                    "type": ["object", "null"],
                                                    "additionalProperties": False,
                                                    "properties": {
                                                        "remedied": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                        "maintained": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                        "introduced": {
                                                            "type": ["object", "null"],
                                                            "properties": {
                                                                "critical": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "high": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "medium": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "low": {
                                                                    "type": ["array", "null"],
                                                                    "items": {"type": "string"},
                                                                },
                                                                "total": {"type": "integer"},
                                                            },
                                                            "required": ["total"],
                                                        },
                                                    },
                                                },
                                            },
                                            "required": [
                                                "fix_version",
                                                "upgrade_type",
                                                "breaking_change",
                                            ],
                                        },
                                    },
                                    "required": [
                                        "closest_min_fix",
                                    ],
                                },
                                "fix_metadata": {
                                    "type": ["object", "null"],
                                    "additionalProperties": False,
                                    "properties": {
                                        "closest_fix_version": {"type": ["string", "null"]},
                                        "upgrade_type": {
                                            "type": ["string", "null"],
                                            "enum": ["unknown", "major", "minor", "patch"],
                                        },
                                        "breaking_change": {"type": "boolean"},
                                        "closest_safe_version": {"type": ["string", "null"]},
                                    },
                                    "required": [
                                        "closest_fix_version",
                                        "upgrade_type",
                                        "breaking_change",
                                    ],
                                },
                                "cve_finding": {"type": ["string", "null"]},
                                "auto_approve": {"type": "boolean"},
                                "upstream_package": {"type": ["string", "null"]},
                                "kev_catalog": {"type": "boolean"},
                            },
                            "required": [
                                "cpes",
                                "description",
                                "id",
                                "severity",
                                "urls",
                                "version_constraint",
                            ],
                        },
                    },
                },
                "required": [
                    "id",
                    "name",
                    "version",
                    "locations",
                    "language",
                    "ecosystem",
                    "package_url",
                    "found_by",
                    "health_metadata",
                    "advisories",
                ],
            },
        },
        "relationships": {
            "type": "array",
            "items": {
                "type": "object",
                "additionalProperties": False,
                "properties": {
                    "from": {"type": "string"},
                    "to": {"type": "array", "items": {"type": "string"}},
                    "type": {"type": "string"},
                },
                "required": ["from", "to"],
            },
        },
    },
    "required": ["sbom_details", "packages", "relationships"],
}



import base64
import logging
import re
from urllib.parse import parse_qs

from labels.model.file import DependencyType, Location, Scope
from labels.model.metadata import Artifact, Digest, HealthMetadata
from labels.model.package_manager import PackageManager
from labels.model.release import Release
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)

PURL_QUALIFIER_DISTRO = "distro"

MALFORMED_PACKAGE_WARNING = (
    "Malformed package. Required fields are missing or data types are incorrect."
)


def purl_qualifiers(
    qualifiers: dict[str, str | None],
    release: Release | None = None,
) -> dict[str, str]:
    if release:
        distro_qualifiers = []
        if release.id_:
            distro_qualifiers.append(release.id_)
        if release.version_id:
            distro_qualifiers.append(release.version_id)
        elif release.build_id:
            distro_qualifiers.append(release.build_id)

        if distro_qualifiers:
            qualifiers[PURL_QUALIFIER_DISTRO] = "-".join(distro_qualifiers)

    return {
        key: qualifiers.get(key, "") or ""
        for key in sorted(qualifiers.keys())
        if qualifiers.get(key)
    }


def _get_distro_params(distro_param: str | None) -> tuple[str | None, str | None]:
    if distro_param is None:
        return None, None

    parts = distro_param.rsplit("-", 1)

    if len(parts) == 2:
        return parts[0], parts[1]

    return parts[0], None


def extract_distro_info(pkg_str: str) -> tuple[str | None, str | None, str | None]:
    parts = pkg_str.split("?", 1)
    if len(parts) != 2:
        return None, None, None

    query = parts[1]
    params = parse_qs(query)
    distro_id, distro_version = _get_distro_params(params.get("distro", [None])[0])
    arch = params.get("arch", [None])[0]
    return distro_id, distro_version, arch


def get_enriched_location(
    base: Location,
    *,
    line: int | None = None,
    is_transitive: bool | None = None,
    is_dev: bool | None = None,
    package_manager: PackageManager = PackageManager.UNKNOWN,
) -> Location:
    update_data: dict[str, object] = {}

    update_data["package_manager"] = package_manager

    if is_dev is not None:
        scope: Scope = Scope.BUILD if is_dev else Scope.RUN
        update_data["scope"] = scope

    if is_transitive is not None:
        dependency_type: DependencyType = (
            DependencyType.TRANSITIVE if is_transitive else DependencyType.DIRECT
        )
        update_data["dependency_type"] = dependency_type

    if line is not None and base.coordinates:
        line_update = {"line": line}
        coordinates = base.coordinates.model_copy(update=line_update)
        update_data["coordinates"] = coordinates

    return base.model_copy(deep=True, update=update_data)


INTEGRITY_PATTERN = re.compile(r"^(sha(?:1|256|384|512))-(.+)\Z")

_HEX_LENGTH_TO_ALGORITHM: dict[int, str] = {
    40: "sha1",
    64: "sha256",
    96: "sha384",
    128: "sha512",
}


def _decode_integrity(digest_str: str) -> tuple[str, str] | None:
    match = INTEGRITY_PATTERN.match(digest_str)
    if not match:
        return None
    _, encoded = match.group(1), match.group(2)
    try:
        hex_value = base64.b64decode(encoded).hex()
    except Exception:  # noqa: BLE001
        return None
    inferred = _HEX_LENGTH_TO_ALGORITHM.get(len(hex_value))
    if not inferred:
        return None
    return inferred, hex_value


def parse_integrity_to_health_metadata(
    integrity_str: str | None,
) -> HealthMetadata | None:
    if not integrity_str:
        return None

    decoded = _decode_integrity(integrity_str.strip())
    if not decoded:
        hex_value = integrity_str.strip()
        if not all(c in "0123456789abcdef" for c in hex_value.lower()):
            return None
        algorithm = _HEX_LENGTH_TO_ALGORITHM.get(len(hex_value))
        if not algorithm:
            return None
        decoded = algorithm, hex_value

    algorithm, hex_value = decoded
    return HealthMetadata(
        artifact=Artifact(
            integrity=Digest(algorithm=algorithm, value=hex_value),
        ),
    )


def log_malformed_package_warning(
    location: Location,
    exception: Exception,
) -> None:
    LOGGER.warning(
        MALFORMED_PACKAGE_WARNING,
        extra={
            "extra": {
                "exception": format_exception(str(exception)),
                "location": location.path(),
            },
        },
    )



import re
from typing import Literal, NamedTuple, NotRequired, TypedDict, TypeGuard

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    add_dependency_to_root_relationship,
    build_relationships,
    create_root_package_from_package_json,
    find_root_location,
    get_dev_dependencies_from_package_json,
    get_direct_dependencies_from_package_json,
    group_packages_by_name,
)
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location, parse_integrity_to_health_metadata

_YARN_OPTIONAL_KEYS = ("checksum", "integrity", "resolution", "resolved")


class YarnPackage(TypedDict):
    line: int
    version: str
    raw_key: NotRequired[str]  # Original key from yarn.lock header
    checksum: NotRequired[str]
    dependencies: NotRequired[list[tuple[str, str]]]
    integrity: NotRequired[str]
    resolution: NotRequired[str]
    resolved: NotRequired[str]


class PackageKey(NamedTuple):
    name: str
    version: str


class ParserState(NamedTuple):
    parsed_yarn_lock: dict[PackageKey, YarnPackage]
    current_package: str | None = None
    current_package_line: int | None = None
    current_package_version: str | None = None
    current_indentation: int | None = None
    current_key: str | None = None
    package_key: PackageKey | None = None
    current_raw_key: str | None = None  # Raw key from header line


def parse_yarn_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    yarn_lock_content = reader.read_closer.read()
    parsed_yarn_lock = _parse_yarn_file(yarn_lock_content)

    resolution_map = _build_resolution_map_from_parsed(parsed_yarn_lock)

    root_package = create_root_package_from_package_json(
        reader.location, package_manager=PackageManager.YARN
    )
    has_package_json = root_package is not None

    direct_dependencies = get_direct_dependencies_from_package_json(reader.location)
    dev_dependencies = get_dev_dependencies_from_package_json(reader.location)

    packages = _extract_packages(
        parsed_yarn_lock,
        reader.location,
        direct_dependencies,
        dev_dependencies,
        has_package_json=has_package_json,
    )

    if root_package:
        packages.append(root_package)

    dependency_map = _build_dependency_map_yarn(parsed_yarn_lock, resolution_map)
    key_map = _build_key_map_yarn(parsed_yarn_lock)
    relationships = build_relationships(packages, dependency_map, reader.location, key_map)

    if root_package:
        root_relationships = _create_root_relationships(
            root_package, packages, direct_dependencies, reader.location
        )
        relationships.extend(root_relationships)

    return packages, relationships


def _parse_yarn_file(yarn_lock_content: str) -> dict[PackageKey, YarnPackage]:
    yarn_lock_lines = yarn_lock_content.strip().split("\n")
    initial_state = ParserState(parsed_yarn_lock={})

    final_state = _process_lines(yarn_lock_lines, initial_state)
    return final_state.parsed_yarn_lock


def _process_lines(lines: list[str], state: ParserState) -> ParserState:
    for index, line in enumerate(lines, 1):
        state = _process_line(line, index, state)
    return state


def _process_line(line: str, index: int, state: ParserState) -> ParserState:
    if not line:
        return state._replace(current_indentation=None)

    if line.startswith("#"):
        return state

    if not line.startswith(" "):
        return _handle_package_header(line, index, state)

    return _process_indented_line(line, state)


def _process_indented_line(line: str, state: ParserState) -> ParserState:
    if state.current_package and state.current_package_line and line.strip().startswith("version"):
        return _handle_version_line(line, state)

    if _is_start_of_list_line(state, line):
        return _handle_list_start(line, state)

    if _is_list_item_line(state, line):
        return _handle_list_item(line, state)

    if state.package_key:
        return _handle_property(line, state)

    return state


def _handle_package_header(line: str, index: int, state: ParserState) -> ParserState:
    current_package, current_package_line = _parse_current_package(line, index)
    raw_key = line.strip().rstrip(":").strip('"')
    return state._replace(
        current_package=current_package,
        current_package_line=current_package_line,
        current_package_version=None,
        package_key=None,
        current_raw_key=raw_key,
    )


def _handle_version_line(line: str, state: ParserState) -> ParserState:
    if not state.current_package or not state.current_package_line:  # pragma: no cover
        return state

    _, raw_version = _resolve_pair(line)
    version = raw_version.strip('"')
    package_key = PackageKey(name=state.current_package, version=version)

    new_package: YarnPackage = {
        "line": state.current_package_line,
        "version": version,
    }
    if state.current_raw_key:
        new_package["raw_key"] = state.current_raw_key

    new_parsed_lock = {**state.parsed_yarn_lock, package_key: new_package}

    return state._replace(
        parsed_yarn_lock=new_parsed_lock,
        current_package_version=version,
        package_key=package_key,
    )


def _handle_list_start(line: str, state: ParserState) -> ParserState:
    if not state.package_key:  # pragma: no cover
        return state

    indentation = _count_indentation(line)
    key = line.strip().split(":")[0]

    if key not in ("dependencies", "optionalDependencies"):
        return state._replace(current_indentation=None, current_key=None)

    existing_deps = state.parsed_yarn_lock[state.package_key].get("dependencies", [])

    updated_package: YarnPackage = {
        **state.parsed_yarn_lock[state.package_key],
        "dependencies": existing_deps if key == "optionalDependencies" else [],
    }
    new_parsed_lock = {**state.parsed_yarn_lock, state.package_key: updated_package}

    return state._replace(
        parsed_yarn_lock=new_parsed_lock,
        current_indentation=indentation,
        current_key=key,
    )


def _handle_list_item(line: str, state: ParserState) -> ParserState:
    if not state.package_key:  # pragma: no cover
        return state

    current_deps = state.parsed_yarn_lock[state.package_key].get("dependencies", [])
    new_dep = _resolve_pair(line)
    new_deps = [*current_deps, new_dep]

    updated_package: YarnPackage = {
        **state.parsed_yarn_lock[state.package_key],
        "dependencies": new_deps,
    }
    new_parsed_lock = {**state.parsed_yarn_lock, state.package_key: updated_package}

    return state._replace(parsed_yarn_lock=new_parsed_lock)


def _is_yarn_optional_key(
    key: str,
) -> TypeGuard[Literal["checksum", "integrity", "resolution", "resolved"]]:
    return key in _YARN_OPTIONAL_KEYS


def _handle_property(line: str, state: ParserState) -> ParserState:
    if not state.package_key:  # pragma: no cover
        return state

    key, value = _resolve_pair(line)
    if _is_yarn_optional_key(key):
        existing_package = state.parsed_yarn_lock[state.package_key]
        updated_package: YarnPackage = {**existing_package}
        updated_package[key] = value.strip('"')
        new_parsed_lock = {**state.parsed_yarn_lock, state.package_key: updated_package}

        return state._replace(parsed_yarn_lock=new_parsed_lock, current_indentation=None)

    return state._replace(current_indentation=None)


def _create_root_relationships(
    root_package: Package,
    packages: list[Package],
    direct_dependencies: set[str],
    location: Location,
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not (location.coordinates and location.coordinates.real_path):
        return relationships

    lockfile_path = location.coordinates.real_path
    root_location = find_root_location(root_package, lockfile_path)

    if not root_location:
        return relationships

    packages_by_name = group_packages_by_name(packages, root_package.id_)

    for dep_name in direct_dependencies:
        for dep_pkg in packages_by_name.get(dep_name, []):
            add_dependency_to_root_relationship(
                dep_pkg, root_package, root_location, lockfile_path, relationships
            )

    return relationships


def _extract_packages(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
    location: Location,
    direct_dependencies: set[str],
    dev_dependencies: set[str],
    *,
    has_package_json: bool,
) -> list[Package]:
    packages = []

    for pkg_info, item in parsed_yarn_lock.items():
        name = _get_name(pkg_info, item)
        version = item.get("version")

        if has_package_json:
            is_direct = name in direct_dependencies
            is_transitive: bool | None = not is_direct

            if is_direct:
                is_dev: bool | None = name in dev_dependencies
            else:
                is_dev = None
        else:
            is_transitive = None
            is_dev = None

        new_location = get_enriched_location(
            location,
            line=item["line"],
            is_transitive=is_transitive,
            is_dev=is_dev,
            package_manager=PackageManager.YARN,
        )

        integrity_str = item.get("integrity") or item.get("checksum")
        health_metadata = parse_integrity_to_health_metadata(integrity_str)

        package = new_simple_npm_package(new_location, name, version, health_metadata)
        if package:
            packages.append(package)

    return packages


def _extract_patch_entry(entry: str) -> tuple[str, str] | None:
    if match := re.search(r"@patch:(.+?)@npm%3A([^#]+)", entry):
        pkg_name = match.group(1)
        version_range_encoded = match.group(2)
        return pkg_name, f"npm:{version_range_encoded}"
    return None


def _extract_npm_entry(entry: str) -> tuple[str, str] | None:
    parts = entry.split("@npm:")
    if len(parts) >= 2:
        return parts[0], f"npm:{parts[-1]}"
    return None


def _extract_scoped_entry(entry: str) -> tuple[str, str] | None:
    last_at = entry.rfind("@")
    pkg_name = entry[:last_at]
    version_range = entry[last_at + 1 :].rstrip(":")
    return pkg_name, version_range


def _extract_simple_entry(entry: str) -> tuple[str, str] | None:
    parts = entry.split("@", 1)
    if len(parts) == 2:
        return parts[0], parts[1].rstrip(":")
    return None


def _process_entry(entry: str, version: str, resolution_map: dict[tuple[str, str], str]) -> None:
    result = None

    if "@patch:" in entry:
        result = _extract_patch_entry(entry)
    elif "@npm:" in entry:
        result = _extract_npm_entry(entry)
    elif entry.startswith("@") and entry.count("@") >= 2:
        result = _extract_scoped_entry(entry)
    elif "@" in entry:
        result = _extract_simple_entry(entry)

    if result:
        pkg_name, version_range = result
        resolution_map[(pkg_name, version_range)] = version


def _build_resolution_map_from_parsed(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
) -> dict[tuple[str, str], str]:
    resolution_map: dict[tuple[str, str], str] = {}

    for item in parsed_yarn_lock.values():
        raw_key = item.get("raw_key")
        version = item.get("version")

        if not raw_key or not version:
            continue

        entries = raw_key.split(", ")
        for raw_entry in entries:
            entry = raw_entry.strip('"')
            _process_entry(entry, version, resolution_map)

    return resolution_map


def _build_dependency_map_yarn(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
    resolution_map: dict[tuple[str, str], str],
) -> DependencyMap:
    dependency_map: DependencyMap = {}

    for pkg_key, item in parsed_yarn_lock.items():
        pkg_name = _get_name(pkg_key, item)
        version = item.get("version")

        if "dependencies" in item and version is not None:
            deps_dict: dict[str, str] = {}

            for dep_name, dep_version in item["dependencies"]:
                clean_name = dep_name.strip('"')
                version_range = dep_version.strip('"')

                resolved_version = resolution_map.get((clean_name, version_range))

                deps_dict[clean_name] = resolved_version or version_range

            package_key = f"{pkg_name}@{version}"
            dependency_map[package_key] = deps_dict

    return dependency_map


def _build_key_map_yarn(
    parsed_yarn_lock: dict[PackageKey, YarnPackage],
) -> dict[tuple[str, int], str]:
    key_map: dict[tuple[str, int], str] = {}

    for pkg_key, item in parsed_yarn_lock.items():
        pkg_name = _get_name(pkg_key, item)
        version = item.get("version")
        line = item.get("line")

        if line is not None and version is not None:
            key_map[(pkg_name, line)] = f"{pkg_name}@{version}"

    return key_map


def _get_name(pkg_info: PackageKey, item: YarnPackage) -> str:
    if resolution := item.get("resolution"):
        is_scoped_package = resolution.startswith("@")
        if is_scoped_package:
            return f"@{resolution.split('@')[1]}"
        return resolution.split("@")[0]

    return pkg_info.name


def _parse_current_package(line: str, index: int) -> tuple[str | None, int | None]:
    line = line.strip()
    if match_ := re.match(r'^"?((?:@\w[\w\-\.]*/)?\w[\w\-\.]*)@', line):
        current_package = match_.groups()[0]
        current_package_line = index
    else:
        current_package = None
        current_package_line = None

    return current_package, current_package_line


def _resolve_pair(line: str) -> tuple[str, str]:
    line = line.strip()
    if ": " in line:
        key, value = line.split(": ")
        return key.strip(), value.strip()

    key, value = line.split(" ", maxsplit=1)
    return key.strip(), value.strip()


def _count_indentation(line: str) -> int:
    # Stripping the leading spaces and comparing the length difference
    return len(line) - len(line.lstrip(" "))


def _is_start_of_list_line(state: ParserState, line: str) -> bool:
    return bool(
        state.current_package and state.current_package_version and line.strip().endswith(":"),
    )


def _is_list_item_line(state: ParserState, line: str) -> bool:
    return bool(
        state.current_package
        and state.current_package_version
        and state.current_key
        and state.current_indentation
        and _count_indentation(line) > state.current_indentation,
    )



import re
from typing import NamedTuple

from gemfileparser import GemfileParser

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.ruby.package_builder import new_gem_package
from labels.parsers.cataloger.ruby.utils import format_ruby_version_constraints
from labels.parsers.cataloger.utils import get_enriched_location


class GemDependency(NamedTuple):
    name: str
    version_constraints: list[str]
    is_development: bool
    line: int


DEPENDENCY_PATTERN = re.compile(
    r"\w+\.add(_runtime|_development)?_dependency\s*\(?\s*"
    r"[\"']([^\"']+)[\"']"
    r"((?:\s*,\s*[\"'][^\"']+[\"'])*)"
    r"\s*\)?"
)

VERSION_CONSTRAINT_PATTERN = re.compile(r"[\"']([^\"']+)[\"']")


def parse_gemspec(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    dependencies = _extract_dependencies_by_line(content)
    packages = _build_packages_from_dependencies(dependencies, reader.location)

    return packages, []


def _extract_dependencies_by_line(content: str) -> list[GemDependency]:
    dependencies: list[GemDependency] = []

    for line_number, line in enumerate(content.splitlines(), start=1):
        preprocessed = GemfileParser.preprocess(line)
        if not preprocessed:
            continue

        dependency = _parse_dependency_line(preprocessed, line_number)
        if dependency:
            dependencies.append(dependency)

    return dependencies


def _parse_dependency_line(line: str, line_number: int) -> GemDependency | None:
    match = DEPENDENCY_PATTERN.search(line)
    if not match:
        return None

    dep_type = str(match.group(1) or "")
    is_development = dep_type == "_development"
    name = match.group(2)
    version_part = match.group(3)

    version_constraints: list[str] = VERSION_CONSTRAINT_PATTERN.findall(version_part)
    if not version_constraints:
        return None

    return GemDependency(name, version_constraints, is_development, line_number)


def _build_packages_from_dependencies(
    dependencies: list[GemDependency], base_location: Location
) -> list[Package]:
    packages: list[Package] = []

    for dep in dependencies:
        version = format_ruby_version_constraints(dep.version_constraints)
        if not version:
            continue

        location = get_enriched_location(
            base_location,
            line=dep.line,
            is_transitive=False,
            is_dev=dep.is_development,
            package_manager=PackageManager.BUNDLER,
        )

        package = new_gem_package(name=dep.name, version=version, location=location)
        if package:
            packages.append(package)

    return packages



import logging
from typing import TYPE_CHECKING

from cyclonedx.model.bom import Bom
from cyclonedx.output import make_outputter
from cyclonedx.schema import OutputFormat as CyclonedxOutputFormat
from cyclonedx.schema import SchemaVersion
from cyclonedx.validation import make_schemabased_validator
from cyclonedx.validation.json import JsonStrictValidator

from labels.model.core import OutputFormat, SbomConfig
from labels.model.package import Package
from labels.model.relationship import Relationship
from labels.model.resolver import Resolver
from labels.model.syft_sbom import SyftSBOM
from labels.output.cyclonedx.file_builder import (
    add_advisories_to_bom,
    add_components_to_bom,
    add_relationships_to_bom,
    create_bom,
    pkg_to_component,
)
from labels.output.utils import set_namespace_version
from labels.utils.exceptions import CycloneDXValidationError

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:  # pragma: no cover
    from cyclonedx.output.json import Json as JsonOutputter
    from cyclonedx.output.xml import Xml as XmlOutputter
    from cyclonedx.validation.xml import XmlValidator


def validate_sbom_xml(serialized_xml: str, file_path: str) -> None:
    xml_validator: XmlValidator = make_schemabased_validator(
        output_format=CyclonedxOutputFormat.XML,
        schema_version=SchemaVersion.V1_6,
    )
    validation_error = xml_validator.validate_str(serialized_xml)

    if validation_error:
        raise CycloneDXValidationError(validation_error)

    LOGGER.info("🆗 Valid CYCLONEDX XML format, generating output file at %s", file_path)


def validate_sbom_json(serialized_json: str, file_path: str) -> None:
    json_validator = JsonStrictValidator(SchemaVersion.V1_6)
    validation_error = json_validator.validate_str(serialized_json)

    if validation_error:
        raise CycloneDXValidationError(validation_error)

    LOGGER.info(
        "🆗 Valid CYCLONEDX JSON format, generating output file at %s",
        file_path,
    )


def format_cyclone_json(bom: Bom, output: str) -> None:
    file_path = f"{output}.json"
    json_output: JsonOutputter = make_outputter(
        bom=bom,
        output_format=CyclonedxOutputFormat.JSON,
        schema_version=SchemaVersion.V1_6,
    )
    serialized_json = json_output.output_as_string()

    validate_sbom_json(serialized_json, file_path)

    json_output.output_to_file(file_path, allow_overwrite=True, indent=2)
    LOGGER.info("✅ Output file successfully generated")


def format_cyclone_xml(bom: Bom, output: str) -> None:
    file_path = f"{output}.xml"
    xml_outputter: XmlOutputter = make_outputter(
        bom=bom,
        output_format=CyclonedxOutputFormat.XML,
        schema_version=SchemaVersion.V1_6,
    )
    serialized_xml = xml_outputter.output_as_string()

    validate_sbom_xml(serialized_xml, file_path)

    xml_outputter.output_to_file(file_path, allow_overwrite=True, indent=2)
    LOGGER.info("✅ Output file successfully generated")


def format_bom_output(bom: Bom, config: SbomConfig) -> None:
    format_handlers = {
        OutputFormat.CYCLONEDX_JSON: format_cyclone_json,
        OutputFormat.CYCLONEDX_XML: format_cyclone_xml,
    }
    handler = format_handlers[config.output_format]
    handler(bom, config.output)


def format_cyclonedx_sbom(
    *,
    packages: list[Package],
    relationships: list[Relationship],
    config: SbomConfig,
    resolver: Resolver | SyftSBOM,
) -> None:
    namespace, version = set_namespace_version(config=config, resolver=resolver)
    bom = create_bom(namespace, version)
    component_cache = {pkg.id_: pkg_to_component(pkg) for pkg in packages}
    add_components_to_bom(bom, component_cache)
    add_advisories_to_bom(bom, packages)
    add_relationships_to_bom(bom, relationships, component_cache)
    format_bom_output(bom, config)



from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.metadata import Artifact, Digest, HealthMetadata
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.python.package_builder import new_python_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter


def parse_pipfile_lock_deps(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    packages = _collect_packages(file_content, reader)

    return packages, []


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue], reader: LocationReadCloser
) -> list[Package]:
    deps: ParsedValue | None = file_content.get("default")
    dev_deps: ParsedValue | None = file_content.get("develop")
    return [
        *_get_packages(reader, deps),
        *_get_packages(reader, dev_deps, is_dev=True),
    ]


def _get_packages(
    reader: LocationReadCloser,
    dependencies: ParsedValue | None,
    *,
    is_dev: bool = False,
) -> list[Package]:
    if dependencies is None or not isinstance(dependencies, IndexedDict):
        return []

    packages = []

    items = dependencies.items()
    for name, value in items:
        if not isinstance(value, IndexedDict) or not isinstance(name, str):
            continue

        version = _get_version(value)

        new_location = get_enriched_location(
            reader.location,
            line=value.position.start.line,
            is_dev=is_dev,
            package_manager=PackageManager.PIPENV,
        )

        package = new_python_package(name=name, version=version, location=new_location)
        if package:
            package.health_metadata = _extract_integrity(value)
            packages.append(package)

    return packages


def _extract_integrity(value: IndexedDict[str, ParsedValue]) -> HealthMetadata | None:
    hashes = value.get("hashes")
    if not isinstance(hashes, IndexedList) or not hashes:
        return None

    first_hash = hashes[0]
    if not isinstance(first_hash, str) or ":" not in first_hash:
        return None

    algorithm, digest_value = first_hash.split(":", 1)
    if not algorithm or not digest_value:
        return None

    return HealthMetadata(
        artifact=Artifact(
            integrity=Digest(algorithm=algorithm, value=digest_value),
        ),
    )


def _get_version(value: IndexedDict[str, ParsedValue]) -> str:
    version = value.get("version")
    if not isinstance(version, str):
        return ""
    return version.strip("=<>~^ ")



import textwrap
from typing import Unpack, cast

import click

from labels.core.configurator import build_labels_config_from_args, build_labels_config_from_file
from labels.core.scanner import execute_labels_scan
from labels.model.core import ScanArgs

SOURCE_CHOICES: list[str] = ["docker", "dir", "docker-daemon", "ecr"]
FORMAT_CHOICES: list[str] = [
    "fluid-json",
    "cyclonedx-json",
    "spdx-json",
    "cyclonedx-xml",
    "spdx-xml",
]


def show_fluid_attacks_banner() -> None:
    logo = textwrap.dedent(
        """
         ───── ⌝
        |    ⌝|  Fluid Attacks
        |  ⌝  |  We hack your software.
         ─────
        """,
    )
    click.secho(logo, fg="red")


@click.group(help="Fluid labels CLI")
def cli() -> None:
    """Entry point for the command-line interface (CLI) of the application.

    This function serves as the main entry point for the CLI, grouping
    all the available commands under a single interface.

    Available commands:
    - config: Load and process a configuration file.
    - scan: Perform a scan based on the provided arguments and options.
    - static-scan: Perform a scan for a full static analysis (SCA and SAST).
    """


@click.command(
    name="config",
    short_help="Load and process a YAML configuration file.",
    epilog="Check out our docs at https://help.fluidattacks.com/portal/en/kb",
)
@click.argument(
    "config_path",
    type=click.Path(exists=True, readable=True, path_type=str),
    required=True,
    metavar="config_path",
)
def run_labels_from_config(config_path: str) -> None:
    """Execute SBOM scan from a configuration file.

    \b
    ARGUMENT:
        config_path:  Path to the YAML configuration file (.yaml or .yml).

    \b
    USAGE EXAMPLES:
        - labels config ./config.yaml   # Load and execute the SBOM scan using config.yaml
        - labels config ./config.yml    # Load and execute the SBOM scan using config.yml

    \b
    The configuration file must be in YAML format (.yaml or .yml).
    """  # noqa: D301
    show_fluid_attacks_banner()
    sbom_config = build_labels_config_from_file(config_path)
    execute_labels_scan(sbom_config)


@click.command(
    name="static_scan",
    short_help="Scan a directory for a static analysis.",
    epilog="Check out our docs at https://help.fluidattacks.com/portal/en/kb",
)
@click.argument(
    "config_path",
    type=click.Path(exists=True, readable=True, path_type=str),
    required=True,
    metavar="config_path",
)
def run_static_scanner(config_path: str) -> None:
    """Execute SBOM scan from a configuration file used for a static analysis (SCA and SAST).

    \b
    ARGUMENT:
        config_path:  Path to the YAML configuration file (.yaml or .yml).

    \b
    USAGE EXAMPLES:
        - labels config ./config.yaml   # Load and execute the SBOM scan using config.yaml

    \b
    The configuration file must be in YAML format (.yaml or .yml).
    """  # noqa: D301
    show_fluid_attacks_banner()
    sbom_config = build_labels_config_from_file(config_path, static_scan=True)
    execute_labels_scan(sbom_config)


@click.command(
    name="scan",
    short_help="Perform an SBOM scan using the specified args.",
    epilog="Check out our docs at https://help.fluidattacks.com/portal/en/kb",
)
@click.option(
    "--source",
    type=cast("click.ParamType", click.Choice(SOURCE_CHOICES, case_sensitive=False)),
    required=True,
    default="dir",
    show_default=True,
    help="Specify the source of the scan: 'docker', 'dir', 'docker-daemon', or 'ecr'.",
)
@click.option(
    "--output",
    default="my_sbom",
    type=click.STRING,
    show_default=True,
    help="Output filename for the scan results.",
)
@click.option(
    "--format",
    type=cast("click.ParamType", click.Choice(FORMAT_CHOICES, case_sensitive=False)),
    default="fluid-json",
    help="Output format for the resulting file.",
)
@click.option(
    "--docker-user",
    default=None,
    type=click.STRING,
    help="Docker registry username, if required.",
)
@click.option(
    "--docker-password",
    default=None,
    type=click.STRING,
    help="Docker registry password, if required.",
)
@click.option(
    "--os",
    default=None,
    type=click.STRING,
    help="Override the OS of the scan default: linux.",
)
@click.option(
    "--arch",
    default=None,
    type=click.STRING,
    help="Override the architecture of the scan default: arm64",
)
@click.option(
    "--aws-external-id",
    default=None,
    type=click.STRING,
    help="AWS external ID for ECR access, if necessary.",
)
@click.option(
    "--aws-role",
    default=None,
    type=click.STRING,
    help="AWS role ARN for ECR access, if necessary.",
)
@click.option(
    "--aws-region",
    default=None,
    type=click.STRING,
    help="AWS region for ECR access. Default: us-east-1.",
)
@click.option(
    "--feature-preview",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable feature preview mode for testing upcoming features.",
)
@click.option(
    "--debug",
    is_flag=True,
    default=False,
    show_default=True,
    help="Enable debug mode for detailed logging.",
)
@click.argument(
    "target",
    type=click.STRING,
    required=True,
)
def run_labels_from_args(target: str, **kwargs: Unpack[ScanArgs]) -> None:
    """Execute SBOM scan with arguments.

    \b
    ARGUMENT:
        target:  The target to scan, depending on the selected source:

    \b
    USAGE EXAMPLES:
        - labels scan my-image:latest --source docker # Scan Docker image
        - labels scan ./my-project --source dir # Scan Local Directory
        - labels scan my-image:latest --source docker-daemon # Scan Local Image
        - labels scan 12345.dkr.ecr.us-east-1.amazonaws.com/my-image --source ecr # Scan ECR Image
    """  # noqa: D301
    show_fluid_attacks_banner()
    sbom_config = build_labels_config_from_args(target, **kwargs)
    execute_labels_scan(sbom_config)


cli.add_command(run_labels_from_config)
cli.add_command(run_labels_from_args)
cli.add_command(run_static_scanner)

if __name__ == "__main__":
    cli(prog_name="labels")



import re
from pathlib import Path

from pydantic import BaseModel

name_and_version_pattern = re.compile(
    r"(?i)^(?P<name>[a-zA-Z][\w.]*?(?:-[a-zA-Z][\w.]*?)*?)"
    r"(?:-(?P<version>\d.*|build\d.*|rc?\d+(?:[^\w].*)?))?$",
)
secondary_version_pattern = re.compile(
    r"(?:[._-](?P<version>(\d.*|build\d+.*|rc?\d+(?:[a-zA-Z].*)?)))?$",
)


class ArchiveFileInfo(BaseModel):
    raw: str
    name: str
    version: str


def parse_file_info(raw: str) -> ArchiveFileInfo:
    cleaned_filename = Path(raw).stem

    matches = name_and_version_pattern.search(cleaned_filename)

    name = _get_subexp(matches, "name")
    version = _get_subexp(matches, "version")

    if not version:
        matches = secondary_version_pattern.search(name)
        secondary_version = _get_subexp(matches, "version")
        if secondary_version:
            name = name[: len(name) - len(secondary_version) - 1]
            version = secondary_version
    return ArchiveFileInfo(raw=raw, name=name, version=version)


def _get_subexp(matches: re.Match[str] | None, subexp_name: str) -> str:
    if matches:
        return matches.group(subexp_name) or ""

    return ""



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.swift.parse_package_resolved import parse_package_resolved
from labels.parsers.cataloger.swift.parse_package_swift import parse_package_swift
from labels.parsers.cataloger.swift.parse_podfile import parse_podfile
from labels.parsers.cataloger.swift.parse_podfile_lock import parse_podfile_lock


def on_next_swift(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in (
                        "**/Package.resolved",
                        "**/.package.resolved",
                        "Package.resolved",
                        ".package.resolved",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_package_resolved,
                            parser_name="parse-swift-parse-package-resolved",
                        ),
                    )
                elif any(fnmatch(value, x) for x in ("**/Podfile", "Podfile")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_podfile,
                            parser_name="parse-swift-parse-podfile",
                        ),
                    )
                elif any(fnmatch(value, x) for x in ("**/Podfile.lock", "Podfile.lock")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_podfile_lock,
                            parser_name="parse-swift-parse-podfile-lock",
                        ),
                    )
                elif any(
                    fnmatch(value, x)
                    for x in (
                        "**/Package.swift",
                        "**/.package.swift",
                        "Package.swift",
                        ".package.swift",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_package_swift,
                            parser_name="parse-swift-parse-package-swift",
                        ),
                    )

            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.php.parse_composer_json import parse_composer_json
from labels.parsers.cataloger.php.parse_composer_lock import parse_composer_lock
from labels.parsers.cataloger.php.parse_installed_json import parse_installed_json
from labels.parsers.cataloger.php.parse_serialized import parse_pecl_serialized


def on_next_php(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, pattern) for pattern in ("**/composer.lock", "composer.lock")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_composer_lock,
                            parser_name="parse-php-composer-lock",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern) for pattern in ("**/composer.json", "composer.json")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_composer_json,
                            parser_name="parse-php-composer-json",
                        ),
                    )
                elif any(
                    fnmatch(value, pattern) for pattern in ("**/installed.json", "installed.json")
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_installed_json,
                            parser_name="parse-php-installed-json",
                        ),
                    )
                elif fnmatch(value, "**/php/.registry/.channel.*/*.reg"):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_pecl_serialized,
                            parser_name="parse-php-pecl-serialized",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



from enum import Enum
from typing import TypedDict

from pydantic import BaseModel

from labels.model.sources import AwsCredentials


class OutputFormat(Enum):
    FLUID_JSON = "fluid-json"
    CYCLONEDX_JSON = "cyclonedx-json"
    CYCLONEDX_XML = "cyclonedx-xml"
    SPDX_JSON = "spdx-json"
    SPDX_XML = "spdx-xml"

    @classmethod
    def from_string(cls: type["OutputFormat"], value: str) -> "OutputFormat":
        for member in cls:
            if member.value == value.lower():
                return member

        error_msg = f"{value} is not a valid {cls.__name__}"
        raise ValueError(error_msg)


class SourceType(Enum):
    DIRECTORY = "dir"
    DOCKER = "docker"
    DOCKER_DAEMON = "docker-daemon"
    ECR = "ecr"
    ECR_WITH_CREDENTIALS = "ecr-with-credentials"

    @classmethod
    def from_string(cls: type["SourceType"], value: str) -> "SourceType":
        for member in cls:
            if member.value == value.lower():
                return member

        error_msg = f"{value} is not a valid {cls.__name__}"
        raise ValueError(error_msg)


class ScanArgs(TypedDict):
    source: str
    format: str
    output: str
    docker_user: str | None
    docker_password: str | None
    os: str | None
    arch: str | None
    aws_external_id: str | None
    aws_role: str | None
    aws_region: str | None
    feature_preview: bool
    debug: bool


class SbomConfig(BaseModel):
    source: str
    source_type: SourceType
    output_format: OutputFormat
    output: str
    execution_id: str | None = None
    include: tuple[str, ...] = (".",)
    exclude: tuple[str, ...] = ()
    docker_user: str | None = None
    docker_password: str | None = None
    os_override: str | None = None
    arch_override: str | None = None
    aws_external_id: str | None = None
    aws_role: str | None = None
    aws_region: str | None = None
    aws_credentials: AwsCredentials | None = None
    debug: bool = False
    feature_preview: bool = False
    include_package_metadata: bool = True
    include_package_advisories: bool = True



import re
import uuid
from pathlib import Path
from urllib.parse import quote, urlunparse

from labels.model.advisories import Advisory
from labels.model.core import SbomConfig, SourceType
from labels.model.package import Ecosystem, Language, Package
from labels.model.relationship import Relationship, RelationshipType
from labels.model.resolver import Resolver
from labels.model.syft_sbom import SyftSBOM
from labels.resolvers.directory import Directory


def sanitize_name(text: str) -> str:
    return re.sub(r"[^a-zA-Z0-9.-]", "-", text)


def get_document_namespace(working_dir: str) -> str:
    input_type = "unknown-source-type"
    path_working_dir = Path(working_dir)

    if path_working_dir.is_file():
        input_type = "file"
    elif path_working_dir.is_dir():
        input_type = "dir"

    unique_id = uuid.uuid4()
    identifier = Path(input_type, str(unique_id))
    if working_dir != ".":
        identifier = Path(input_type, f"{working_dir}-{unique_id}")

    return urlunparse(
        ("https", "fluidattacks.com", str(identifier), "", "", ""),
    )


def get_relative_path(absolute_path: str) -> str:
    root_path = Path(absolute_path)
    current_dir = Path.cwd()
    try:
        return str(root_path.relative_to(current_dir))
    except ValueError:
        return absolute_path


def set_namespace_version(
    config: SbomConfig, resolver: Resolver | SyftSBOM
) -> tuple[str, str | None]:
    if config.source_type == SourceType.DIRECTORY and isinstance(resolver, Directory):
        return get_relative_path(resolver.root), None

    if (
        config.source_type
        in {
            SourceType.DOCKER,
            SourceType.ECR,
            SourceType.DOCKER_DAEMON,
            SourceType.ECR_WITH_CREDENTIALS,
        }
        and isinstance(resolver, SyftSBOM)
        and resolver.source
    ):
        return resolver.source.namespace, resolver.source.version

    return "", None


def create_packages_for_test() -> tuple[list[Package], list[Relationship]]:
    test_from_pkg = Package(
        name="source-package",
        version="1.0.0",
        locations=[],
        language=Language.PYTHON,
        ecosystem=Ecosystem.PYPI,
        p_url="pkg:test/source@1.0.0",
        ecosystem_data=None,
        found_by="python-pyproject-toml-cataloger",
        advisories=[
            Advisory(
                id="test-advisory",
                source="https://test-url.com",
                severity_level="HIGH",
                epss=0.75,
                percentile=8.5,
                details="Test advisory description",
                vulnerable_version=">=1.0.0",
                cpes=[],
                ecosystem="github",
                severity_v4=None,
            ),
        ],
    )
    test_to_pkg = Package(
        name="target-package",
        version="1.0.0",
        locations=[],
        language=Language.PYTHON,
        ecosystem=Ecosystem.PYPI,
        p_url="pkg:test/target@1.0.0",
        ecosystem_data=None,
        found_by="python-pyproject-toml-cataloger",
    )
    test_packages = [
        Package(
            name="test-package",
            version="1.0.0",
            locations=[],
            language=Language.PYTHON,
            ecosystem=Ecosystem.PYPI,
            p_url="pkg:python/test-package@1.0.0",
            ecosystem_data=None,
            found_by="python-pyproject-toml-cataloger",
        ),
        test_from_pkg,
        test_to_pkg,
    ]
    test_relationships = [
        Relationship(
            from_=test_from_pkg.id_,
            to_=test_to_pkg.id_,
            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
        ),
    ]
    return test_packages, test_relationships


def sanitize_spdx_path(text: str) -> str:
    return quote(text, safe="/-._~:")



from labels.model.indexables import IndexedDict, ParsedValue
from labels.model.package import Ecosystem, Package
from labels.model.relationship import Relationship, RelationshipType
from labels.utils.strings import normalize_name


def build_relationships_from_json_dependencies(
    dependencies: IndexedDict[str, ParsedValue] | IndexedDict[ParsedValue, ParsedValue],
    current_package: Package,
    packages_by_key: dict[tuple[str, str], Package],
) -> list[Relationship]:
    relationships: list[Relationship] = []

    for dependency_name, dependency_version in dependencies.items():
        if not isinstance(dependency_name, str) or not isinstance(dependency_version, str):
            continue

        dependency_name_normalized = normalize_name(dependency_name, Ecosystem.NUGET)
        dependency_package = packages_by_key.get((dependency_name_normalized, dependency_version))

        if dependency_package:
            relationships.append(
                Relationship(
                    from_=dependency_package.id_,
                    to_=current_package.id_,
                    type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                )
            )

    return relationships



from typing import NamedTuple

from tree_sitter import Node

from labels.model.file import Location, LocationReadCloser
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.swift.package_builder import new_cocoa_pods_package
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.ruby import find_method_calls, parse_ruby_with_tree_sitter


class PodDeclaration(NamedTuple):
    name: str
    version: str
    line: int


class CollectedArguments(NamedTuple):
    pod_args: list[str]
    git_url: str | None
    tag: str | None
    first_line: int


def parse_podfile(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    root_node = parse_ruby_with_tree_sitter(content)
    if not root_node:
        return [], []

    pod_declarations = _extract_pod_declarations(root_node)
    if not pod_declarations:
        return [], []

    packages = _build_packages(pod_declarations, reader.location)

    return packages, []


def _extract_pod_declarations(root_node: Node) -> list[PodDeclaration]:
    pod_declarations: list[PodDeclaration] = []

    pod_calls = find_method_calls(root_node, "pod")

    for pod_call in pod_calls:
        name, version, line = _extract_pod_arguments(pod_call)

        if not name or not version:
            continue

        pod_declarations.append(
            PodDeclaration(
                name=name,
                version=version,
                line=line,
            )
        )

    return pod_declarations


def _extract_pod_arguments(node: Node) -> tuple[str | None, str, int]:
    args_node = node.child_by_field_name("arguments")
    if not args_node:
        return None, "", 0

    collected = _collect_argument_data(args_node)

    if collected.git_url and collected.tag:
        return (
            collected.git_url,
            collected.tag,
            collected.first_line,
        )

    if collected.pod_args:
        return (
            collected.pod_args[0],
            " ".join(collected.pod_args[1:]),
            collected.first_line,
        )

    return None, "", 0


def _collect_argument_data(args_node: Node) -> CollectedArguments:
    pod_args: list[str] = []
    git_url: str | None = None
    tag: str | None = None
    first_line: int = 0

    for child in args_node.children:
        if child.type == "string":
            arg_value, line = _extract_string_value(child)
            if arg_value:
                pod_args.append(arg_value)
                if first_line == 0:
                    first_line = line
        elif child.type == "pair":
            key, value = _extract_from_pair(child)
            if key == ":tag":
                tag = value
            if key == ":git":
                git_url = value

    return CollectedArguments(
        pod_args=pod_args,
        git_url=git_url,
        tag=tag,
        first_line=first_line,
    )


def _extract_string_value(string_node: Node) -> tuple[str | None, int]:
    content_node = string_node.child(1)
    if not content_node or content_node.type != "string_content" or not content_node.text:
        return None, 0

    value = content_node.text.decode("utf-8")
    line = string_node.start_point[0] + 1 if string_node.start_point else 0
    return value, line


def _extract_from_pair(pair_node: Node) -> tuple[str | None, str | None]:
    if len(pair_node.children) < 3:
        return None, None

    key_node = pair_node.children[0]
    key_text = key_node.text.decode("utf-8") if key_node.text else ""
    if key_node.type != "simple_symbol":
        return None, None

    value_node = pair_node.children[2]
    if value_node.type != "string":
        return None, None

    string_value, _ = _extract_string_value(value_node)
    if not string_value:
        return None, None

    if key_text in [":tag", ":git"]:
        return key_text, string_value

    return None, None


def _build_packages(pod_declarations: list[PodDeclaration], location: Location) -> list[Package]:
    packages: list[Package] = []

    for pod_decl in pod_declarations:
        enriched_location = get_enriched_location(
            location,
            is_transitive=False,
            line=pod_decl.line,
            package_manager=PackageManager.COCOAPODS,
        )

        package = new_cocoa_pods_package(
            name=pod_decl.name,
            version=pod_decl.version,
            location=enriched_location,
        )

        if package:
            packages.append(package)

    return packages



from pydantic import BaseModel

from labels.model.constraints import NonEmptyStr


class SyftAnnotation(BaseModel):  # pragma: no cover
    evidence: str


class SyftLocation(BaseModel):  # pragma: no cover
    path: NonEmptyStr
    access_path: NonEmptyStr
    layer_id: NonEmptyStr
    annotations: SyftAnnotation | None = None


class SyftProject(BaseModel):  # pragma: no cover
    group_id: NonEmptyStr | None = None
    artifact_id: NonEmptyStr | None = None
    version: NonEmptyStr | None = None
    name: NonEmptyStr | None = None
    parent: dict[NonEmptyStr, NonEmptyStr] | None = None


class SyftManifestItem(BaseModel):  # pragma: no cover
    key: str = ""
    value: str = ""


class SyftManifest(BaseModel):  # pragma: no cover
    main: list[SyftManifestItem] | None = None
    sections: list[list[SyftManifestItem]] | None = None


class SyftMetadata(BaseModel):  # pragma: no cover
    package: NonEmptyStr
    version: NonEmptyStr
    provides: list[NonEmptyStr] = []
    dependencies: list[NonEmptyStr] = []
    maintainer: str = ""
    origin_package: str = ""
    architecture: str = ""
    source: str = ""
    source_version: str = ""
    pre_dependencies: list[NonEmptyStr] = []
    name: str = ""
    epoch: int | None = None
    arch: str = ""
    release: str = ""
    source_rpm: str = ""
    base_package: str = ""
    packager: str = ""
    requires_dist: list[NonEmptyStr] = []
    provides_extra: list[NonEmptyStr] = []
    manifest: SyftManifest | None = None
    pom_properties: dict[NonEmptyStr, str] = {}
    pom_project: SyftProject | None = None


class SyftArtifact(BaseModel):  # pragma: no cover
    id_: str
    name: str
    version: str
    type: str
    locations: list[SyftLocation]
    metadata: SyftMetadata
    p_url: str
    found_by: str
    language: str


class SyftSource(BaseModel):  # pragma: no cover
    namespace: NonEmptyStr
    version: str


class SyftRelationship(BaseModel):  # pragma: no cover
    parent: NonEmptyStr
    child: NonEmptyStr
    type: NonEmptyStr


class SyftSBOM(BaseModel):  # pragma: no cover
    source: SyftSource
    artifacts: list[SyftArtifact]
    relationships: list[SyftRelationship]



from labels.model.package import Package


def count_vulns_by_severity(packages: list[Package]) -> dict[str, int]:
    counts = {"critical": 0, "high": 0, "medium": 0, "low": 0}
    for pkg in packages:
        advisories = pkg.advisories or []
        for adv in advisories:
            sev_lower = adv.severity_level.lower()
            if sev_lower in counts:
                counts[sev_lower] += 1
    return counts



import json
import logging
from typing import Any

from labels.model.package import Package
from labels.model.relationship import Relationship

LOGGER = logging.getLogger(__name__)


def serialize_packages(packages: list[Package]) -> list[dict[str, Any]]:  # type: ignore[explicit-any]
    raw_pkgs = [pkg.serialize() for pkg in packages]
    return validate_pkgs(raw_pkgs)


def validate_pkgs(raw_sbom_pkgs: list[dict[str, Any]]) -> list[dict[str, Any]]:  # type: ignore[explicit-any]
    pkgs_to_remove: list[int] = []
    for index, raw_pkg in enumerate(raw_sbom_pkgs):
        try:
            json.dumps(raw_pkg)
        except TypeError as err:
            LOGGER.warning(
                "Cannot serialize package at index %d: %s. Error: %s",
                index,
                raw_pkg,
                str(err),
            )
            pkgs_to_remove.append(index)

    for index in sorted(pkgs_to_remove, reverse=True):
        raw_sbom_pkgs.pop(index)
    return raw_sbom_pkgs


def build_relationship_map(
    relationships: list[Relationship],
) -> list[dict[str, str | list[str]]]:
    grouped: dict[tuple[str, str], set[str]] = {}

    for rel in relationships:
        if rel.type.value == "dependency-of":
            # Invert relationship and change type to be semantically correct
            # Input: dep_pkg DEPENDENCY_OF package (dep_pkg is a dependency OF package)
            # Output: package DEPENDS_ON dep_pkg (package depends on dep_pkg)
            key = (rel.to_, "depends-on")
            grouped.setdefault(key, set()).add(rel.from_)
        elif rel.type.value == "described-by":
            # from describes to (lock → non-lock)
            key = (rel.from_, rel.type.value)
            grouped.setdefault(key, set()).add(rel.to_)
    result: list[dict[str, str | list[str]]] = [
        {"from": from_ref, "to": list(deps), "type": rel_type}
        for (from_ref, rel_type), deps in grouped.items()
    ]

    return result


def build_sbom_metadata(
    namespace: str, version: str | None, timestamp: str
) -> dict[str, str | None]:
    return {
        "name": namespace,
        "version": version,
        "timestamp": timestamp,
        "tool": "Fluid-Labels",
        "organization": "Fluid attacks",
    }



import base64
import json
import logging
import os
import re
import shutil
import subprocess
import tempfile
from collections.abc import Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any

from labels.model.syft_sbom import (
    SyftAnnotation,
    SyftArtifact,
    SyftLocation,
    SyftMetadata,
    SyftRelationship,
    SyftSBOM,
    SyftSource,
)
from labels.utils.exceptions import (
    DockerImageNotFoundError,
    InvalidImageReferenceError,
    SyftNotFoundError,
    SyftOutputParsingError,
)

LOGGER = logging.getLogger(__name__)

ADDITIONAL_CATALOGERS = [
    "dart-pubspec-lock-cataloger",
    "dart-pubspec-cataloger",
    "swift-package-manager-cataloger",
    "cocoapods-cataloger",
]


def _get_syft_path() -> str:
    syft_path = shutil.which("syft")
    if not syft_path:
        raise SyftNotFoundError
    return syft_path


def _build_platform_arg(
    *,
    os_override: str | None = None,
    arch_override: str | None = None,
) -> list[str]:
    os_value = os_override or "linux"
    arch_value = arch_override or "arm64"
    return ["--platform", f"{os_value}/{arch_value}"]


def _get_formatted_image(*, image_ref: str, daemon: bool = False) -> str:
    image_ref_pattern = (
        r"^(?:(?P<host>[\w\.\-]+(?:\:\d+)?)/)?"
        r"(?P<namespace>(?:[\w\.\-]+(?:/[\w\.\-]+)*)?/)?"
        r"(?P<image>[\w\.\-]+)(?::(?P<tag>[\w\.\-]+))?(?:@"
        r"(?P<digest>sha256:[A-Fa-f0-9]{64}))?$"
    )

    syft_prefix = "docker" if daemon else "registry"
    prefixes = {"registry": "docker://", "docker": "docker-daemon:"}
    for prefix, prefix_value in prefixes.items():
        if image_ref.startswith(prefix_value):
            image_ref = image_ref.replace(prefix_value, "", 1)
            syft_prefix = prefix
            break

    if re.match(image_ref_pattern, image_ref):
        return f"{syft_prefix}:{image_ref}"

    raise InvalidImageReferenceError(image_ref)


@contextmanager
def _syft_auth_env(
    *,
    image_ref: str | None = None,
    username: str | None = None,
    password: str | None = None,
    aws_creds: str | None = None,
) -> Iterator[dict[str, str]]:
    if username and password:
        LOGGER.debug("Setting SYFT_REGISTRY_AUTH credentials for authenticated scan")
        yield {
            "SYFT_REGISTRY_AUTH_USERNAME": username,
            "SYFT_REGISTRY_AUTH_PASSWORD": password,
        }
        return

    if image_ref and aws_creds:
        config_dir = tempfile.mkdtemp(prefix=".syft")
        LOGGER.debug("Creating temporary Docker config dir at %s", config_dir)

        try:
            docker_config_path = Path(config_dir, "config.json")
            registry_url = image_ref.split("/", 1)[0]

            encoded_auth = base64.b64encode(aws_creds.encode("utf-8")).decode("utf-8")
            docker_config = {
                "auths": {
                    registry_url: {
                        "auth": encoded_auth,
                    },
                }
            }

            with docker_config_path.open("w", encoding="utf-8") as f:
                json.dump(docker_config, f)
            yield {
                "DOCKER_CONFIG": config_dir,
            }
        finally:
            LOGGER.debug("Cleaning up temporary DOCKER_CONFIG directory: %s", config_dir)
            shutil.rmtree(config_dir, ignore_errors=True)

        return

    yield {}


def _get_metadata(metadata: dict[str, Any]) -> SyftMetadata:  # type: ignore[explicit-any]
    dependencies = (
        metadata.get("depends", [])
        if metadata.get("depends")
        else metadata.get("pullDependencies", [])
    )
    release = metadata.get("release", "") if isinstance(metadata.get("release"), str) else ""
    package = str(metadata.get("package")) if metadata.get("package") else "UNKNOWN"
    version = str(metadata.get("version")) if metadata.get("version") else "UNKNOWN"
    source = metadata.get("source", "") if isinstance(metadata.get("source"), str) else ""

    return SyftMetadata(
        package=package,
        version=version,
        provides=metadata.get("provides", []),
        dependencies=dependencies,
        maintainer=metadata.get("maintainer", ""),
        origin_package=metadata.get("originPackage", ""),
        architecture=metadata.get("architecture", ""),
        source=source,
        source_version=metadata.get("sourceVersion", ""),
        pre_dependencies=metadata.get("preDepends", []),
        name=metadata.get("name", ""),
        epoch=metadata.get("epoch"),
        release=release,
        source_rpm=metadata.get("sourceRpm", ""),
        base_package=metadata.get("basepackage", ""),
        packager=metadata.get("packager", ""),
        requires_dist=metadata.get("requiresDist", []),
        provides_extra=metadata.get("providesExtra", []),
        manifest=metadata.get("manifest"),
        pom_properties=metadata.get("pomProperties", {}),
        pom_project=metadata.get("pomProject"),
    )


def _get_syft_source(source: dict[str, Any]) -> SyftSource:  # type: ignore[explicit-any]
    namespace = str(source.get("metadata", {}).get("userInput"))
    version = ""

    if (digests := source.get("metadata", {}).get("repoDigests")) and (first_digest := digests[0]):
        version = str(first_digest).split("@", 1)[1]

    return SyftSource(
        namespace=namespace,
        version=version,
    )


def _get_syft_relationships(relationships: list[dict[str, Any]]) -> list[SyftRelationship]:  # type: ignore[explicit-any]
    return [
        SyftRelationship(
            parent=parent,
            child=child,
            type=type_,
        )
        for relationship in relationships
        if (parent := relationship.get("parent"))
        and (child := relationship.get("child"))
        and (type_ := relationship.get("type"))
    ]


def _get_syft_artifacts(artifacts: list[dict[str, Any]]) -> list[SyftArtifact]:  # type: ignore[explicit-any]
    return [
        SyftArtifact(
            id_=str(artifact.get("id")),
            name=str(artifact.get("name")),
            version=str(artifact.get("version")),
            type=str(artifact.get("type")),
            locations=[
                SyftLocation(
                    path=location.get("path"),
                    access_path=location.get("path"),
                    layer_id=location.get("layerID"),
                    annotations=(
                        SyftAnnotation(
                            evidence=location.get("annotations", {}).get("evidence", ""),
                        )
                        if location.get("annotations")
                        else None
                    ),
                )
                for location in artifact.get("locations", [])
            ],
            p_url=str(artifact.get("purl")),
            metadata=_get_metadata(artifact.get("metadata", {})),
            found_by=str(artifact.get("foundBy")),
            language=str(artifact.get("language")),
        )
        for artifact in artifacts
        if artifact.get("id")
        and artifact.get("name")
        and artifact.get("version")
        and artifact.get("version") != "UNKNOWN"
    ]


def _parse_syft_output(raw: dict[str, Any], image_ref: str) -> SyftSBOM:  # type: ignore[explicit-any]
    if "artifacts" not in raw:
        raise SyftOutputParsingError(image_ref)

    return SyftSBOM(
        artifacts=_get_syft_artifacts(raw["artifacts"]),
        source=_get_syft_source(raw.get("source", {})),
        relationships=_get_syft_relationships(raw.get("artifactRelationships", [])),
    )


def get_docker_sbom(  # noqa: PLR0913
    image_ref: str,
    *,
    username: str | None = None,
    password: str | None = None,
    os_override: str | None = None,
    arch_override: str | None = None,
    aws_creds: str | None = None,
    daemon: bool = False,
) -> SyftSBOM:
    syft_path = _get_syft_path()
    formatted_img = _get_formatted_image(image_ref=image_ref, daemon=daemon)

    LOGGER.info("Retrieving SBOM for image: %s", image_ref)

    command_args = [
        syft_path,
        "scan",
        formatted_img,
        *_build_platform_arg(os_override=os_override, arch_override=arch_override),
        "--select-catalogers",
        ",".join("+" + cataloger for cataloger in ADDITIONAL_CATALOGERS),
        "-o",
        "json",
    ]

    with _syft_auth_env(
        image_ref=image_ref,
        username=username,
        password=password,
        aws_creds=aws_creds,
    ) as syft_envs:
        try:
            result = subprocess.run(  # noqa: S603
                command_args,
                check=True,
                capture_output=True,
                text=True,
                env={**os.environ, **syft_envs},
            )
        except subprocess.CalledProcessError as error:
            platform = f"{os_override or 'linux'}/{arch_override or 'arm64'}"
            LOGGER.exception(
                "Failed to scan image %s for platform %s. "
                "Ensure the image exists and use --os and --arch "
                "to specify a different platform.",
                image_ref,
                platform,
            )
            raise DockerImageNotFoundError(image_ref, error.stderr.strip()) from error

    return _parse_syft_output(json.loads(result.stdout), image_ref)



from fnmatch import fnmatch

import reactivex
from reactivex.abc import ObserverBase, SchedulerBase

from labels.model.parser import Request
from labels.parsers.cataloger.ruby.parse_gemfile import parse_gemfile
from labels.parsers.cataloger.ruby.parse_gemfile_lock import parse_gemfile_lock
from labels.parsers.cataloger.ruby.parse_gemspec import parse_gemspec


def on_next_ruby(
    source: reactivex.Observable[str],
) -> reactivex.Observable[Request]:
    def subscribe(
        observer: ObserverBase[Request],
        scheduler: SchedulerBase | None = None,
    ) -> reactivex.abc.DisposableBase:
        def on_next(value: str) -> None:
            try:
                if any(
                    fnmatch(value, x)
                    for x in (
                        "**/Gemfile.lock",
                        "Gemfile.lock",
                        "**/gems.locked",
                        "gems.locked",
                    )
                ):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_gemfile_lock,
                            parser_name="parse-gemfile-lock",
                        ),
                    )
                elif any(fnmatch(value, x) for x in ("**/Gemfile", "Gemfile")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_gemfile,
                            parser_name="parse-gemfile",
                        ),
                    )
                elif any(fnmatch(value, x) for x in ("**/*.gemspec", "*.gemspec")):
                    observer.on_next(
                        Request(
                            real_path=value,
                            parser=parse_gemspec,
                            parser_name="parse-gemspec",
                        ),
                    )
            except Exception as ex:  # noqa: BLE001
                observer.on_error(ex)

        return source.subscribe(
            on_next,
            observer.on_error,
            observer.on_completed,
            scheduler=scheduler,
        )

    return reactivex.create(subscribe)



import re

from labels.model.file import Location, LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.javascript.common import (
    DependencyMap,
    add_dependency_to_root_relationship,
    build_relationships,
    create_root_package_from_package_json,
    find_root_location,
    get_dev_dependencies_from_package_json,
    get_direct_dependencies_from_package_json,
    group_packages_by_name,
)
from labels.parsers.cataloger.javascript.package_builder import new_simple_npm_package
from labels.parsers.cataloger.utils import get_enriched_location, parse_integrity_to_health_metadata
from labels.parsers.collection.json import parse_json_with_tree_sitter

VERSION_PATTERN = re.compile(r"^(.+)@(\d+\.\d+\.\d+(?:-[\w.]+)?)$")


def parse_bun_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    content = reader.read_closer.read()
    file_content = parse_json_with_tree_sitter(_strip_jsonc_comments(content))

    if not isinstance(file_content, IndexedDict):
        return [], []

    direct_dependencies = get_direct_dependencies_from_package_json(reader.location)
    dev_dependencies = get_dev_dependencies_from_package_json(reader.location)
    root_package = create_root_package_from_package_json(
        reader.location, package_manager=PackageManager.BUN
    )

    packages = _collect_packages(
        reader.location,
        file_content,
        direct_dependencies,
        dev_dependencies,
    )

    if root_package:
        packages.append(root_package)

    dependency_map = _build_dependency_map(file_content)
    key_map = _build_key_map(file_content)

    relationships = build_relationships(packages, dependency_map, reader.location, key_map)

    if root_package:
        root_relationships = _create_root_relationships(
            root_package, packages, direct_dependencies, reader.location
        )
        relationships.extend(root_relationships)

    return packages, relationships


def _strip_jsonc_comments(content: str) -> str:
    lines = content.split("\n")
    result = []
    for line in lines:
        stripped = line.lstrip()
        if stripped.startswith("//"):
            continue
        result.append(line)
    return "\n".join(result)


def _collect_packages(
    location: Location,
    file_content: IndexedDict[str, ParsedValue],
    direct_dependencies: set[str],
    dev_dependencies: set[str],
) -> list[Package]:
    packages: list[Package] = []

    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return packages

    for package_key, package_entry in packages_dict.items():
        if not isinstance(package_entry, IndexedList):
            continue

        name_version = _parse_package_entry(package_entry)
        if not name_version:
            continue

        name, version = name_version

        is_direct = name in direct_dependencies
        is_transitive = not is_direct
        is_dev = name in dev_dependencies if is_direct else None

        position = packages_dict.get_key_position(package_key)
        new_location = get_enriched_location(
            location,
            line=position.start.line,
            is_transitive=is_transitive,
            is_dev=is_dev,
            package_manager=PackageManager.BUN,
        )

        integrity_str = package_entry[3] if len(package_entry) > 3 else None
        health_metadata = parse_integrity_to_health_metadata(
            str(integrity_str) if isinstance(integrity_str, str) else None
        )

        package = new_simple_npm_package(new_location, name, version, health_metadata)
        if package:
            packages.append(package)

    return packages


def _parse_package_entry(entry: IndexedList[ParsedValue]) -> tuple[str, str] | None:
    if len(entry) < 1:
        return None

    first_element = entry[0]
    if not isinstance(first_element, str):
        return None

    match = VERSION_PATTERN.match(first_element)
    if not match:
        return None

    return match.group(1), match.group(2)


def _build_dependency_map(file_content: IndexedDict[str, ParsedValue]) -> DependencyMap:
    dependency_map: DependencyMap = {}

    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return dependency_map

    for package_entry in packages_dict.values():
        result = _extract_package_dependencies(package_entry)
        if result:
            map_key, deps = result
            dependency_map[map_key] = deps

    return dependency_map


def _extract_package_dependencies(
    package_entry: ParsedValue,
) -> tuple[str, dict[str, str]] | None:
    if not isinstance(package_entry, IndexedList):
        return None

    name_version = _parse_package_entry(package_entry)
    if not name_version or len(package_entry) < 3:
        return None

    name, version = name_version
    deps_obj = package_entry[2]
    if not isinstance(deps_obj, IndexedDict):
        return None

    dependencies_dict = deps_obj.get("dependencies")
    if not isinstance(dependencies_dict, IndexedDict):
        return None

    deps = {
        dep_name: _extract_version(dep_version)
        for dep_name, dep_version in dependencies_dict.items()
        if isinstance(dep_version, str)
    }

    if not deps:
        return None

    return f"{name}@{version}", deps


def _build_key_map(
    file_content: IndexedDict[str, ParsedValue],
) -> dict[tuple[str, int], str]:
    key_map: dict[tuple[str, int], str] = {}

    packages_dict = file_content.get("packages")
    if not isinstance(packages_dict, IndexedDict):
        return key_map

    for package_key, package_entry in packages_dict.items():
        if not isinstance(package_entry, IndexedList):
            continue

        name_version = _parse_package_entry(package_entry)
        if not name_version:
            continue

        name, version = name_version
        position = packages_dict.get_key_position(package_key)
        key_map[(name, position.start.line)] = f"{name}@{version}"

    return key_map


def _extract_version(version_spec: str) -> str:
    match = re.match(r"^(\d+\.\d+\.\d+(?:-[\w.]+)?)", version_spec)
    if match:
        return match.group(1)
    return version_spec


def _create_root_relationships(
    root_package: Package,
    packages: list[Package],
    direct_dependencies: set[str],
    location: Location,
) -> list[Relationship]:
    relationships: list[Relationship] = []

    if not (location.coordinates and location.coordinates.real_path):
        return relationships

    lockfile_path = location.coordinates.real_path
    root_location = find_root_location(root_package, lockfile_path)

    if not root_location:
        return relationships

    packages_by_name = group_packages_by_name(packages, root_package.id_)

    for dep_name in direct_dependencies:
        for dep_pkg in packages_by_name.get(dep_name, []):
            add_dependency_to_root_relationship(
                dep_pkg, root_package, root_location, lockfile_path, relationships
            )

    return relationships



import asyncio

from labels.integrations.docker import get_docker_sbom
from labels.integrations.ecr import ecr_connection, get_token
from labels.model.core import SbomConfig, SourceType
from labels.model.sources import AwsCredentials, AwsRole
from labels.model.syft_sbom import SyftSBOM
from labels.resolvers.directory import Directory
from labels.utils.exceptions import (
    AwsCredentialsNotFoundError,
    AwsExternalIdNotDefinedError,
    AwsRoleNotDefinedError,
    UnexpectedSBOMSourceError,
)


def _resolve_ecr_credentials(role: AwsRole, region: str) -> str:
    return asyncio.run(ecr_connection(role, region=region))


def _resolve_ecr_token(aws_credentials: AwsCredentials, region: str) -> str:
    return asyncio.run(get_token(aws_credentials, region=region))


def _handle_directory(sbom_config: SbomConfig) -> Directory:
    return Directory(
        root=sbom_config.source,
        include=sbom_config.include,
        exclude=sbom_config.exclude,
    )


def _handle_docker(*, sbom_config: SbomConfig, is_daemon: bool = False) -> SyftSBOM:
    return get_docker_sbom(
        sbom_config.source,
        username=sbom_config.docker_user,
        password=sbom_config.docker_password,
        os_override=sbom_config.os_override,
        arch_override=sbom_config.arch_override,
        daemon=is_daemon,
    )


def _handle_ecr(sbom_config: SbomConfig) -> SyftSBOM:
    if not sbom_config.aws_role:
        raise AwsRoleNotDefinedError

    if not sbom_config.aws_external_id:
        raise AwsExternalIdNotDefinedError

    role = AwsRole(
        external_id=sbom_config.aws_external_id,
        role=sbom_config.aws_role,
    )

    region = sbom_config.aws_region or "us-east-1"

    aws_creds = _resolve_ecr_credentials(role, region=region)
    return get_docker_sbom(
        sbom_config.source,
        aws_creds=aws_creds,
        os_override=sbom_config.os_override,
        arch_override=sbom_config.arch_override,
    )


def handle_ecr_with_credentials(
    image_uri: str,
    aws_credentials: AwsCredentials,
    *,
    os_override: str | None = None,
    arch_override: str | None = None,
    aws_region: str | None = None,
) -> SyftSBOM:
    region = aws_region or "us-east-1"
    token = _resolve_ecr_token(aws_credentials, region=region)
    return get_docker_sbom(
        image_uri,
        aws_creds=f"AWS:{token}",
        os_override=os_override,
        arch_override=arch_override,
    )


def resolve_sbom_source(sbom_config: SbomConfig) -> Directory | SyftSBOM:
    match sbom_config.source_type:
        case SourceType.DIRECTORY:
            return _handle_directory(sbom_config)
        case SourceType.DOCKER:
            return _handle_docker(sbom_config=sbom_config)
        case SourceType.DOCKER_DAEMON:
            return _handle_docker(sbom_config=sbom_config, is_daemon=True)
        case SourceType.ECR:
            return _handle_ecr(sbom_config)
        case SourceType.ECR_WITH_CREDENTIALS:
            if not sbom_config.aws_credentials:
                raise AwsCredentialsNotFoundError

            return handle_ecr_with_credentials(
                image_uri=sbom_config.source,
                aws_credentials=sbom_config.aws_credentials,
                os_override=sbom_config.os_override,
                arch_override=sbom_config.arch_override,
                aws_region=sbom_config.aws_region,
            )
        case _:
            raise UnexpectedSBOMSourceError(sbom_config.source_type)



from io import TextIOWrapper

from labels.model.file import Coordinates, DependencyType, Location, LocationReadCloser, Scope


def new_location(real_path: str) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
        ),
        access_path=real_path,
    )


def new_default_location_from_path(real_path: str, line: int | None = None) -> Location:
    return Location(
        coordinates=Coordinates(
            real_path=real_path,
            line=line,
        ),
        access_path=real_path,
        scope=Scope.UNDETERMINABLE,
        dependency_type=DependencyType.UNDETERMINABLE,
        reachable_cves=[],
    )


def new_default_location_read_closer(
    str_path: str, read_closer: TextIOWrapper
) -> LocationReadCloser:
    location = new_default_location_from_path(str_path)
    return LocationReadCloser(location=location, read_closer=read_closer)



from typing import NamedTuple

from labels.model.ecosystem_data.base import EcosystemDataModel


class JavaPomParent(NamedTuple):
    group_id: str
    artifact_id: str
    version: str


class JavaPomProject(NamedTuple):
    group_id: str | None = None
    artifact_id: str | None = None
    version: str | None = None
    name: str | None = None
    parent: JavaPomParent | None = None


class JavaManifest(NamedTuple):
    main: dict[str, str]
    sections: list[dict[str, str]] | None = None


class JavaPomProperties(NamedTuple):
    name: str | None = None
    group_id: str | None = None
    artifact_id: str | None = None
    version: str | None = None


class JavaArchive(EcosystemDataModel):
    manifest: JavaManifest | None = None
    pom_properties: JavaPomProperties | None = None
    pom_project: JavaPomProject | None = None



import json
import logging

from cvss import CVSS4
from cvss.exceptions import CVSS4MalformedError
from pydantic import BaseModel, ValidationError

from labels.model.advisories import Advisory, AdvisoryRecord
from labels.utils.strings import format_exception

LOGGER = logging.getLogger(__name__)


def _parse_json_list(
    raw: str | None,
    default: list[str] | None = None,
) -> list[str] | None:
    if raw:
        result: list[str] = json.loads(raw)
        return result
    return default


class DatabasesToDownload(BaseModel):
    roots_db: bool
    images_db: bool


def _resolve_severity_level(
    severity_v4: str | None, precalculated_severity_level: str | None
) -> str:
    return _get_severity_level_from_cvss_v4(severity_v4) or precalculated_severity_level or "Low"


def _get_severity_level_from_cvss_v4(severity_v4: str | None) -> str | None:
    if not severity_v4:
        return None

    try:
        return str(CVSS4(severity_v4).severity)
    except CVSS4MalformedError:
        LOGGER.exception(
            "Could not determine severity level from provided CVSS4 vector: %s", severity_v4
        )
        return None


def generate_cpe(ecosystem: str, package_name: str, vulnerable_version: str) -> str:
    part = "a"
    vendor = package_name.split(":")[0] if ":" in package_name else "*"
    product = package_name.lower()
    version = vulnerable_version
    language = ecosystem
    update = edition = sw_edition = target_sw = target_hw = other = "*"

    return (
        f"cpe:2.3:{part}:{vendor}:{product}:{version}:{update}:{edition}:"
        f"{language}:{sw_edition}:{target_sw}:{target_hw}:{other}"
    )


def create_advisory_from_record(
    advisory_db_record: AdvisoryRecord,
    ecosystem: str,
    package_name: str,
    version: str,
    upstream_package: str | None = None,
) -> Advisory | None:
    try:
        return Advisory(
            id=advisory_db_record[0],
            source=advisory_db_record[1],
            vulnerable_version=advisory_db_record[2],
            severity_level=_resolve_severity_level(
                severity_v4=advisory_db_record[4],
                precalculated_severity_level=advisory_db_record[3],
            ),
            severity_v4=advisory_db_record[4],
            epss=float(advisory_db_record[5]) if advisory_db_record[5] else 0.0,
            details=advisory_db_record[6],
            percentile=float(advisory_db_record[7]) if advisory_db_record[7] else 0.0,
            cwe_ids=_parse_json_list(advisory_db_record[8], default=["CWE-1395"]),
            cve_finding=advisory_db_record[9],
            auto_approve=bool(advisory_db_record[10]),
            fixed_versions=_parse_json_list(advisory_db_record[11]),
            cpes=[generate_cpe(ecosystem, package_name, version)],
            ecosystem=ecosystem,
            upstream_package=upstream_package if upstream_package else None,
            kev_catalog=bool(advisory_db_record[12]),
            platform_version=(
                advisory_db_record[13]
                if len(advisory_db_record) > 13 and advisory_db_record[13]
                else None
            ),
        )
    except ValidationError as ex:
        LOGGER.exception(
            "Unable to build advisory from database record",
            extra={
                "exception": format_exception(str(ex)),
                "advisory_db_record": advisory_db_record,
            },
        )
        return None



from typing import cast

from labels.model.file import LocationReadCloser
from labels.model.indexables import IndexedDict, IndexedList, ParsedValue
from labels.model.package import Package
from labels.model.package_manager import PackageManager
from labels.model.relationship import Relationship, RelationshipType
from labels.model.release import Environment
from labels.model.resolver import Resolver
from labels.parsers.cataloger.php.package_builder import new_package_from_composer
from labels.parsers.cataloger.utils import get_enriched_location
from labels.parsers.collection.json import parse_json_with_tree_sitter

EMPTY_DICT: IndexedDict[str, ParsedValue] = IndexedDict()


def parse_composer_lock(
    _resolver: Resolver | None,
    _environment: Environment | None,
    reader: LocationReadCloser,
) -> tuple[list[Package], list[Relationship]]:
    file_content = parse_json_with_tree_sitter(reader.read_closer.read())
    if not isinstance(file_content, IndexedDict):
        return [], []

    packages = _collect_packages(file_content, reader)
    relationships = _collect_relationships(file_content, packages)

    return packages, relationships


def _collect_packages(
    file_content: IndexedDict[str, ParsedValue], reader: LocationReadCloser
) -> list[Package]:
    packages: list[Package] = []

    for is_dev, raw_package in _get_all_packages_with_dev_flag(file_content):
        name = str(raw_package.get("name", "")) or None
        version = str(raw_package.get("version", "")) or None

        if not name:
            continue

        new_location = get_enriched_location(
            reader.location,
            line=raw_package.get_key_position("name").start.line,
            is_dev=is_dev,
            is_transitive=False,
            package_manager=PackageManager.COMPOSER,
        )

        package = new_package_from_composer(name=name, version=version, location=new_location)
        if package:
            packages.append(package)

    return packages


def _collect_relationships(
    package_json: IndexedDict[str, ParsedValue], packages: list[Package]
) -> list[Relationship]:
    relationships: list[Relationship] = []

    all_packages = [package for _, package in _get_all_packages_with_dev_flag(package_json)]

    for raw_package in all_packages:
        name = str(raw_package.get("name", "")) or None
        if not name:
            continue

        parsed_package = next((x for x in packages if x.name == name), None)
        if not parsed_package:
            continue

        if isinstance(raw_package.get("require"), IndexedDict):
            require = cast("dict[str, str]", raw_package.get("require"))
            deps = list(require.keys())

            for dep_name in deps:
                package_dep = next((x for x in packages if x.name == dep_name), None)
                if package_dep:
                    relationships.append(
                        Relationship(
                            from_=parsed_package.id_,
                            to_=package_dep.id_,
                            type=RelationshipType.DEPENDENCY_OF_RELATIONSHIP,
                        ),
                    )

    return relationships


def _get_all_packages_with_dev_flag(
    package_json: IndexedDict[str, ParsedValue],
) -> list[tuple[bool, IndexedDict[str, ParsedValue]]]:
    packages: list[tuple[bool, IndexedDict[str, ParsedValue]]] = []

    production_packages = package_json.get("packages")
    development_packages = package_json.get("packages-dev")

    if isinstance(production_packages, IndexedList):
        packages.extend(
            [
                (False, package)
                for package in production_packages
                if isinstance(package, IndexedDict)
            ]
        )

    if isinstance(development_packages, IndexedList):
        packages.extend(
            [
                (True, package)
                for package in development_packages
                if isinstance(package, IndexedDict)
            ]
        )

    return packages



from enum import Enum


class Algorithm(Enum):
    SHA1 = "sha1"
    SHA224 = "sha224"
    SHA256 = "sha256"
    SHA384 = "sha384"
    SHA512 = "sha512"
    SHA3_256 = "sha3-256"
    SHA3_384 = "sha3-384"
    SHA3_512 = "sha3-512"
    BLAKE2B_256 = "blake2b-256"
    BLAKE2B_384 = "blake2b-384"
    BLAKE2B_512 = "blake2b-512"
    BLAKE3 = "blake3"
    MD2 = "md2"
    MD4 = "md4"
    MD5 = "md5"
    ADLER32 = "adler32"
    MD6 = "md6"


algorithm_length: dict[Algorithm, str] = {
    Algorithm.SHA1: "40",
    Algorithm.SHA224: "56",
    Algorithm.SHA256: "64",
    Algorithm.SHA384: "96",
    Algorithm.SHA512: "128",
    Algorithm.SHA3_256: "64",
    Algorithm.SHA3_384: "96",
    Algorithm.SHA3_512: "128",
    Algorithm.BLAKE2B_256: "64",
    Algorithm.BLAKE2B_384: "96",
    Algorithm.BLAKE2B_512: "128",
    Algorithm.BLAKE3: "256",
    Algorithm.MD2: "32",
    Algorithm.MD4: "32",
    Algorithm.MD5: "32",
    Algorithm.MD6: "512",
    Algorithm.ADLER32: "8",
}


def infer_algorithm(digest_value: str | None) -> str | None:
    if digest_value:
        for algorithm, length in algorithm_length.items():
            if len(digest_value) == int(length):
                return algorithm.value
    return None
